@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -29,12 +29,9 @@
 
 #include "config.h"
 
-#include <inttypes.h>
-#include <math.h>
-#include <stdio.h>
-#include <stdlib.h>
-
 #include "cache.h"
+#include "cache_director.h"
+#include "cache_filter.h"
 #include "hash/hash_slinger.h"
 #include "vcl.h"
 #include "vtim.h"
@@ -45,17 +42,18 @@
  * XXX: Should this be merged over there ?
  */
 
-static struct object *
-vbf_allocobj(struct busyobj *bo, unsigned l, uint16_t nhttp)
+static int
+vbf_allocobj(struct busyobj *bo, unsigned l)
 {
-	struct object *obj;
+	struct objcore *oc;
 	const char *storage_hint;
 	double lifetime;
 
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
+	oc = bo->fetch_objcore;
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 
-	lifetime = bo->exp.ttl + bo->exp.grace + bo->exp.keep;
+	lifetime = oc->exp.ttl + oc->exp.grace + oc->exp.keep;
 
 	if (bo->uncacheable || lifetime < cache_param->shortlived)
 		storage_hint = TRANSIENT_STORAGE;
@@ -64,25 +62,23 @@ vbf_allocobj(struct busyobj *bo, unsigned l, uint16_t nhttp)
 
 	bo->storage_hint = NULL;
 
-	obj = STV_NewObject(bo, storage_hint, l, nhttp);
-
-	if (obj != NULL)
-		return (obj);
+	if (STV_NewObject(bo->fetch_objcore, bo->wrk, storage_hint, l))
+		return (1);
 
 	if (storage_hint != NULL && !strcmp(storage_hint, TRANSIENT_STORAGE))
-		return (NULL);
+		return (0);
 
 	/*
 	 * Try to salvage the transaction by allocating a shortlived object
 	 * on Transient storage.
 	 */
 
-	if (bo->exp.ttl > cache_param->shortlived)
-		bo->exp.ttl = cache_param->shortlived;
-	bo->exp.grace = 0.0;
-	bo->exp.keep = 0.0;
-	obj = STV_NewObject(bo, TRANSIENT_STORAGE, l, nhttp);
-	return (obj);
+	if (oc->exp.ttl > cache_param->shortlived)
+		oc->exp.ttl = cache_param->shortlived;
+	oc->exp.grace = 0.0;
+	oc->exp.keep = 0.0;
+	return (STV_NewObject(bo->fetch_objcore, bo->wrk,
+	    TRANSIENT_STORAGE, l));
 }
 
 /*--------------------------------------------------------------------
@@ -92,13 +88,11 @@ vbf_allocobj(struct busyobj *bo, unsigned l, uint16_t nhttp)
 static int
 vbf_beresp2obj(struct busyobj *bo)
 {
-	unsigned l;
-	char *b;
+	unsigned l, l2;
+	const char *b;
+	uint8_t *bp;
 	struct vsb *vary = NULL;
 	int varyl = 0;
-	uint16_t nhttp;
-	struct object *obj;
-	struct http *hp, *hp2;
 
 	l = 0;
 
@@ -108,7 +102,7 @@ vbf_beresp2obj(struct busyobj *bo)
 		if (varyl > 0) {
 			AN(vary);
 			assert(varyl == VSB_len(vary));
-			l += varyl;
+			l += PRNDUP((intptr_t)varyl);
 		} else if (varyl < 0) {
 			/*
 			 * Vary parse error
@@ -124,48 +118,39 @@ vbf_beresp2obj(struct busyobj *bo)
 			AZ(vary);
 	}
 
-	l += http_EstimateWS(bo->beresp,
-	    bo->uncacheable ? HTTPH_R_PASS : HTTPH_A_INS, &nhttp);
+	l2 = http_EstimateWS(bo->beresp,
+	    bo->uncacheable ? HTTPH_R_PASS : HTTPH_A_INS);
+	l += l2;
 
 	if (bo->uncacheable)
 		bo->fetch_objcore->flags |= OC_F_PASS;
 
-	obj = vbf_allocobj(bo, l, nhttp);
-
-	if (obj == NULL)
+	if (!vbf_allocobj(bo, l))
 		return (-1);
 
-	CHECK_OBJ_NOTNULL(obj, OBJECT_MAGIC);
-
-	AZ(bo->fetch_obj);
-	bo->fetch_obj = obj;
-
 	if (vary != NULL) {
-		obj->vary = (void *)WS_Copy(obj->http->ws,
-		    VSB_data(vary), varyl);
-		AN(obj->vary);
-		(void)VRY_Validate(obj->vary);
+		b = ObjSetattr(bo->wrk, bo->fetch_objcore, OA_VARY, varyl,
+		    VSB_data(vary));
 		VSB_delete(vary);
 	}
 
-	obj->vxid = bo->vsl->wid;
-	WS_Assert(bo->ws_o);
+	AZ(ObjSetU32(bo->wrk, bo->fetch_objcore, OA_VXID, VXID(bo->vsl->wid)));
+
+	/* for HTTP_Encode() VSLH call */
+	bo->beresp->logtag = SLT_ObjMethod;
 
 	/* Filter into object */
-	hp = bo->beresp;
-	hp2 = obj->http;
+	bp = ObjSetattr(bo->wrk, bo->fetch_objcore, OA_HEADERS, l2, NULL);
+	AN(bp);
+	HTTP_Encode(bo->beresp, bp, l2,
+	    bo->uncacheable ? HTTPH_R_PASS : HTTPH_A_INS);
 
-	hp2->logtag = SLT_ObjMethod;
-	http_FilterResp(hp, hp2, bo->uncacheable ? HTTPH_R_PASS : HTTPH_A_INS);
-	http_CopyHome(hp2);
-
-	if (http_GetHdr(hp, H_Last_Modified, &b))
-		obj->last_modified = VTIM_parse(b);
+	if (http_GetHdr(bo->beresp, H_Last_Modified, &b))
+		AZ(ObjSetDouble(bo->wrk, bo->fetch_objcore, OA_LASTMODIFIED,
+		    VTIM_parse(b)));
 	else
-		obj->last_modified = floor(bo->exp.t_origin);
-
-	/* Disassociate the obj from the bo's workspace */
-	hp2->ws = NULL;
+		AZ(ObjSetDouble(bo->wrk, bo->fetch_objcore, OA_LASTMODIFIED,
+		    floor(bo->fetch_objcore->exp.t_origin)));
 
 	return (0);
 }
@@ -177,15 +162,13 @@ vbf_beresp2obj(struct busyobj *bo)
 static enum fetch_step
 vbf_stp_mkbereq(const struct worker *wrk, struct busyobj *bo)
 {
-	char *p;
+	const char *q;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 	CHECK_OBJ_NOTNULL(bo->req, REQ_MAGIC);
 
 	assert(bo->state == BOS_INVALID);
-	AZ(bo->vbc);
-	AZ(bo->should_close);
 	AZ(bo->storage_hint);
 
 	HTTP_Setup(bo->bereq0, bo->ws, bo->vsl, SLT_BereqMethod);
@@ -200,17 +183,19 @@ vbf_stp_mkbereq(const struct worker *wrk, struct busyobj *bo)
 		AN(bo->req);
 		bo->req = NULL;
 		http_CopyHome(bo->bereq0);
-	}
+	} else
+		AZ(bo->stale_oc);
 
-	if (bo->ims_obj != NULL && bo->ims_obj->http->status == 200) {
-		if (http_GetHdr(bo->ims_obj->http, H_Last_Modified, &p)) {
+	if (bo->stale_oc != NULL &&
+	    ObjCheckFlag(bo->wrk, bo->stale_oc, OF_IMSCAND)) {
+		q = HTTP_GetHdrPack(bo->wrk, bo->stale_oc, H_Last_Modified);
+		if (q != NULL)
 			http_PrintfHeader(bo->bereq0,
-			    "If-Modified-Since: %s", p);
-		}
-		if (http_GetHdr(bo->ims_obj->http, H_ETag, &p)) {
+			    "If-Modified-Since: %s", q);
+		q = HTTP_GetHdrPack(bo->wrk, bo->stale_oc, H_ETag);
+		if (q != NULL)
 			http_PrintfHeader(bo->bereq0,
-			    "If-None-Match: %s", p);
-		}
+			    "If-None-Match: %s", q);
 	}
 
 	HTTP_Setup(bo->bereq, bo->ws, bo->vsl, SLT_BereqMethod);
@@ -229,41 +214,34 @@ vbf_stp_mkbereq(const struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_retry(struct worker *wrk, struct busyobj *bo)
 {
-	unsigned owid, wid;
+	struct vfp_ctx *vfc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	vfc = bo->vfc;
+	CHECK_OBJ_NOTNULL(vfc, VFP_CTX_MAGIC);
 
 	assert(bo->state == BOS_REQ_DONE);
 
 	VSLb_ts_busyobj(bo, "Retry", W_TIM_real(wrk));
 
-	/* VDI_CloseFd must have been called before */
-	AZ(bo->vbc);
-	bo->should_close = 0;
+	/* VDI_Finish must have been called before */
+	assert(bo->director_state == DIR_S_NULL);
 
 	/* reset other bo attributes - See VBO_GetBusyObj */
 	bo->storage_hint = NULL;
 	bo->do_esi = 0;
-	bo->do_gzip = 0;
-	bo->do_gunzip = 0;
 	bo->do_stream = 1;
-	bo->uncacheable = 0;
-	bo->abandon = 0;
 
 	/* reset fetch processors */
-	bo->failed = 0;
-	bo->vfp_nxt = 0;
+	vfc->failed = 0;
+	VFP_Close(vfc);
+	VFP_Setup(vfc);
 
 	// XXX: BereqEnd + BereqAcct ?
-	wid = VXID_Get(&wrk->vxid_pool);
-	VSLb(bo->vsl, SLT_Link, "bereq %u retry", wid);
-	VSLb(bo->vsl, SLT_End, "%s", "");
-	VSL_Flush(bo->vsl, 0);
-	owid = bo->vsl->wid & VSL_IDENTMASK;
-	bo->vsl->wid = wid | VSL_BACKENDMARKER;
-	VSLb(bo->vsl, SLT_Begin, "bereq %u retry", owid);
+	VSL_ChgId(bo->vsl, "bereq", "retry", VXID_Get(wrk, VSL_BACKENDMARKER));
 	VSLb_ts_busyobj(bo, "Start", bo->t_prev);
+	http_VSL_log(bo->bereq);
 
 	return (F_STP_STARTFETCH);
 }
@@ -275,15 +253,12 @@ vbf_stp_retry(struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 {
-	int i, do_ims;
+	int i;
 	double now;
-	char time_str[VTIM_FORMAT_SIZE];
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 
-	AZ(bo->vbc);
-	AZ(bo->should_close);
 	AZ(bo->storage_hint);
 
 	if (bo->do_pass)
@@ -291,11 +266,12 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 	else
 		AZ(bo->req);
 
-	http_PrintfHeader(bo->bereq,
-	    "X-Varnish: %u", bo->vsl->wid & VSL_IDENTMASK);
+	if (bo->retries > 0)
+		http_Unset(bo->bereq, "\012X-Varnish:");
 
+	http_PrintfHeader(bo->bereq, "X-Varnish: %u", VXID(bo->vsl->wid));
 
-	VCL_backend_fetch_method(bo->vcl, wrk, NULL, bo, bo->bereq->ws);
+	VCL_backend_fetch_method(bo->vcl, wrk, NULL, bo, NULL);
 
 	bo->uncacheable = bo->do_pass;
 	if (wrk->handling == VCL_RET_ABANDON)
@@ -307,26 +283,17 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 
 	assert(bo->state <= BOS_REQ_DONE);
 
-	i = V1F_fetch_hdr(wrk, bo, bo->req);
-	/*
-	 * If we recycle a backend connection, there is a finite chance
-	 * that the backend closed it before we get a request to it.
-	 * Do a single retry in that case.
-	 */
-	if (i == 1) {
-		VSLb_ts_busyobj(bo, "Beresp", W_TIM_real(wrk));
-		VSC_C_main->backend_retry++;
-		i = V1F_fetch_hdr(wrk, bo, bo->req);
-	}
+	AZ(bo->htc);
+	i = VDI_GetHdr(wrk, bo);
+
 	now = W_TIM_real(wrk);
 	VSLb_ts_busyobj(bo, "Beresp", now);
 
 	if (i) {
-		AZ(bo->vbc);
+		assert(bo->director_state == DIR_S_NULL);
 		return (F_STP_ERROR);
 	}
 
-	AN(bo->vbc);
 	http_VSL_log(bo->beresp);
 
 	if (!http_GetHdr(bo->beresp, H_Date, NULL)) {
@@ -341,8 +308,7 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 		 *
 		 * If we didn't get a Date header, we assign one here.
 		 */
-		VTIM_format(now, time_str);
-		http_PrintfHeader(bo->beresp, "Date: %s", time_str);
+		http_TimeHeader(bo->beresp, "Date: ", now);
 	}
 
 	/*
@@ -356,56 +322,132 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 	/*
 	 * Figure out how the fetch is supposed to happen, before the
 	 * headers are adultered by VCL
-	 * NB: Also sets other wrk variables
 	 */
-	bo->htc.body_status = RFC2616_Body(bo, &wrk->stats);
+	if (!strcasecmp(http_GetMethod(bo->bereq), "head")) {
+		/*
+		 * A HEAD request can never have a body in the reply,
+		 * no matter what the headers might say.
+		 * [RFC7231 4.3.2 p25]
+		 */
+		wrk->stats->fetch_head++;
+		bo->htc->body_status = BS_NONE;
+	} else if (http_GetStatus(bo->beresp) <= 199) {
+		/*
+		 * 1xx responses never have a body.
+		 * [RFC7230 3.3.2 p31]
+		 * ... but we should never see them.
+		 */
+		wrk->stats->fetch_1xx++;
+		bo->htc->body_status = BS_ERROR;
+	} else if (http_IsStatus(bo->beresp, 204)) {
+		/*
+		 * 204 is "No Content", obviously don't expect a body.
+		 * [RFC7230 3.3.1 p29 and 3.3.2 p31]
+		 */
+		wrk->stats->fetch_204++;
+		if ((http_GetHdr(bo->beresp, H_Content_Length, NULL) &&
+		    bo->htc->content_length != 0) ||
+		    http_GetHdr(bo->beresp, H_Transfer_Encoding, NULL))
+			bo->htc->body_status = BS_ERROR;
+		else
+			bo->htc->body_status = BS_NONE;
+	} else if (http_IsStatus(bo->beresp, 304)) {
+		/*
+		 * 304 is "Not Modified" it has no body.
+		 * [RFC7230 3.3 p28]
+		 */
+		wrk->stats->fetch_304++;
+		bo->htc->body_status = BS_NONE;
+	} else if (bo->htc->body_status == BS_CHUNKED) {
+		wrk->stats->fetch_chunked++;
+	} else if (bo->htc->body_status == BS_LENGTH) {
+		assert(bo->htc->content_length > 0);
+		wrk->stats->fetch_length++;
+	} else if (bo->htc->body_status == BS_EOF) {
+		wrk->stats->fetch_eof++;
+	} else if (bo->htc->body_status == BS_ERROR) {
+		wrk->stats->fetch_bad++;
+	} else if (bo->htc->body_status == BS_NONE) {
+		wrk->stats->fetch_none++;
+	} else {
+		WRONG("wrong bodystatus");
+	}
 
-	if (bo->htc.body_status == BS_ERROR) {
-		AN (bo->vbc);
-		VDI_CloseFd(&bo->vbc, &bo->acct);
-		VSLb(bo->vsl, SLT_VCL_Error, "Body cannot be fetched");
+	if (bo->htc->body_status == BS_ERROR) {
+		bo->htc->doclose = SC_RX_BODY;
+		VDI_Finish(bo->wrk, bo);
+		VSLb(bo->vsl, SLT_Error, "Body cannot be fetched");
+		assert(bo->director_state == DIR_S_NULL);
 		return (F_STP_ERROR);
 	}
 
 	/*
 	 * What does RFC2616 think about TTL ?
 	 */
-	EXP_Clr(&bo->exp);
+	EXP_Clr(&bo->fetch_objcore->exp);
 	RFC2616_Ttl(bo, now);
 
 	/* private objects have negative TTL */
 	if (bo->fetch_objcore->flags & OC_F_PRIVATE)
-		bo->exp.ttl = -1.;
+		bo->fetch_objcore->exp.ttl = -1.;
 
 	AZ(bo->do_esi);
+	AZ(bo->was_304);
 
-	if (bo->ims_obj != NULL && bo->ims_obj->http->status == 200 &&
-	    (http_GetHdr(bo->ims_obj->http, H_Last_Modified, NULL) ||
-	     http_GetHdr(bo->ims_obj->http, H_ETag, NULL)) &&
-	    bo->beresp->status == 304) {
-		http_Unset(bo->beresp, H_Content_Length);
-		http_Merge(bo->ims_obj->http, bo->beresp,
-		    bo->ims_obj->changed_gzip);
-		assert(bo->beresp->status == 200);
-		do_ims = 1;
-	} else
-		do_ims = 0;
+	if (http_IsStatus(bo->beresp, 304)) {
+		if (bo->stale_oc != NULL &&
+		    ObjCheckFlag(bo->wrk, bo->stale_oc, OF_IMSCAND)) {
+			if (ObjCheckFlag(bo->wrk, bo->stale_oc, OF_CHGGZIP)) {
+				/*
+				 * If we changed the gzip status of the object
+				 * the stored Content_Encoding controls we
+				 * must weaken any new ETag we get.
+				 */
+				http_Unset(bo->beresp, H_Content_Encoding);
+				RFC2616_Weaken_Etag(bo->beresp);
+			}
+			http_Unset(bo->beresp, H_Content_Length);
+			HTTP_Merge(bo->wrk, bo->stale_oc, bo->beresp);
+			assert(http_IsStatus(bo->beresp, 200));
+			bo->was_304 = 1;
+		} else if (!bo->do_pass) {
+			/*
+			 * Backend sent unallowed 304
+			 */
+			VSLb(bo->vsl, SLT_Error,
+			    "304 response but not conditional fetch");
+			bo->htc->doclose = SC_RX_BAD;
+			VDI_Finish(bo->wrk, bo);
+			return (F_STP_FAIL);
+		}
+	}
 
-	VCL_backend_response_method(bo->vcl, wrk, NULL, bo, bo->beresp->ws);
+	bo->vfc->bo = bo;
+	bo->vfc->oc = bo->fetch_objcore;
+	bo->vfc->wrk = bo->wrk;
+	bo->vfc->http = bo->beresp;
+	bo->vfc->esi_req = bo->bereq;
 
-	if (wrk->handling == VCL_RET_ABANDON)
+	VCL_backend_response_method(bo->vcl, wrk, NULL, bo, NULL);
+
+	if (wrk->handling == VCL_RET_ABANDON) {
+		bo->htc->doclose = SC_RESP_CLOSE;
+		VDI_Finish(bo->wrk, bo);
 		return (F_STP_FAIL);
+	}
 
 	if (wrk->handling == VCL_RET_RETRY) {
-		AN (bo->vbc);
-		VDI_CloseFd(&bo->vbc, &bo->acct);
-		bo->should_close = 0;
-		bo->retries++;
-		if (bo->retries <= cache_param->max_retries)
+		if (bo->htc->body_status != BS_NONE)
+			bo->htc->doclose = SC_RESP_CLOSE;
+		if (bo->director_state != DIR_S_NULL)
+			VDI_Finish(bo->wrk, bo);
+
+		if (bo->retries++ < cache_param->max_retries)
 			return (F_STP_RETRY);
 
 		VSLb(bo->vsl, SLT_VCL_Error,
 		    "Too many retries, delivering 503");
+		assert(bo->director_state == DIR_S_NULL);
 		return (F_STP_ERROR);
 	}
 
@@ -418,7 +460,75 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 
 	assert(wrk->handling == VCL_RET_DELIVER);
 
-	return (do_ims ? F_STP_CONDFETCH : F_STP_FETCH);
+	return (bo->was_304 ? F_STP_CONDFETCH : F_STP_FETCH);
+}
+
+/*--------------------------------------------------------------------
+ */
+
+static void
+vbf_fetch_body_helper(struct busyobj *bo)
+{
+	ssize_t l;
+	uint8_t *ptr;
+	enum vfp_status vfps = VFP_ERROR;
+	ssize_t est;
+	struct vfp_ctx *vfc;
+
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	vfc = bo->vfc;
+	CHECK_OBJ_NOTNULL(vfc, VFP_CTX_MAGIC);
+
+	AN(vfc->vfp_nxt);
+
+	est = bo->htc->content_length;
+	if (est < 0)
+		est = 0;
+
+	do {
+		if (bo->abandon) {
+			/*
+			 * A pass object and delivery was terminated
+			 * We don't fail the fetch, in order for hit-for-pass
+			 * objects to be created.
+			 */
+			AN(vfc->oc->flags & OC_F_PASS);
+			VSLb(vfc->wrk->vsl, SLT_FetchError,
+			    "Pass delivery abandoned");
+			vfps = VFP_END;
+			bo->htc->doclose = SC_RX_BODY;
+			break;
+		}
+		AZ(vfc->failed);
+		l = est;
+		assert(l >= 0);
+		if (VFP_GetStorage(vfc, &l, &ptr) != VFP_OK) {
+			bo->htc->doclose = SC_RX_BODY;
+			break;
+		}
+
+		AZ(vfc->failed);
+		vfps = VFP_Suck(vfc, ptr, &l);
+		if (l > 0 && vfps != VFP_ERROR) {
+			bo->acct.beresp_bodybytes += l;
+			VBO_extend(bo, l);
+			if (est >= l)
+				est -= l;
+			else
+				est = 0;
+		}
+	} while (vfps == VFP_OK);
+
+	VFP_Close(vfc);
+
+	if (vfps == VFP_ERROR) {
+		AN(vfc->failed);
+		(void)VFP_Error(vfc, "Fetch pipeline failed to process");
+		bo->htc->doclose = SC_RX_BODY;
+	}
+
+	if (!bo->do_stream)
+		ObjTrimStore(bo->wrk, vfc->oc);
 }
 
 /*--------------------------------------------------------------------
@@ -427,11 +537,11 @@ vbf_stp_startfetch(struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_fetch(struct worker *wrk, struct busyobj *bo)
 {
-	struct object *obj;
-	ssize_t est;
+	const char *p;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
 
 	assert(wrk->handling == VCL_RET_DELIVER);
 
@@ -446,158 +556,137 @@ vbf_stp_fetch(struct worker *wrk, struct busyobj *bo)
 	 *	no Content-Encoding		--> object is not gzip'ed.
 	 *	anything else			--> do nothing wrt gzip
 	 *
-	 * XXX: BS_NONE/cl==0 should avoid gzip/gunzip
 	 */
 
 	/* We do nothing unless the param is set */
 	if (!cache_param->http_gzip_support)
 		bo->do_gzip = bo->do_gunzip = 0;
 
-	bo->is_gzip = http_HdrIs(bo->beresp, H_Content_Encoding, "gzip");
+	if (bo->htc->content_length == 0)
+		http_Unset(bo->beresp, H_Content_Encoding);
 
-	bo->is_gunzip = !http_GetHdr(bo->beresp, H_Content_Encoding, NULL);
+	if (bo->htc->body_status != BS_NONE) {
+		bo->is_gzip =
+		    http_HdrIs(bo->beresp, H_Content_Encoding, "gzip");
+		bo->is_gunzip =
+		    !http_GetHdr(bo->beresp, H_Content_Encoding, NULL);
+		assert(bo->is_gzip == 0 || bo->is_gunzip == 0);
+	}
 
-	/* It can't be both */
-	assert(bo->is_gzip == 0 || bo->is_gunzip == 0);
-
-	/* We won't gunzip unless it is gzip'ed */
-	if (bo->do_gunzip && !bo->is_gzip)
+	/* We won't gunzip unless it is non-empty and gzip'ed */
+	if (bo->htc->body_status == BS_NONE ||
+	    bo->htc->content_length == 0 ||
+	    (bo->do_gunzip && !bo->is_gzip))
 		bo->do_gunzip = 0;
 
-	/* We wont gzip unless it is ungziped */
-	if (bo->do_gzip && !bo->is_gunzip)
+	/* We wont gzip unless it is non-empty and ungzip'ed */
+	if (bo->htc->body_status == BS_NONE ||
+	    bo->htc->content_length == 0 ||
+	    (bo->do_gzip && !bo->is_gunzip))
 		bo->do_gzip = 0;
 
-	AN(bo->vbc);
-	est = V1F_Setup_Fetch(bo);
-
-	if (est == 0) {
-		/*
-		 * If the length is known to be zero, it's not gziped.
-		 * A similar issue exists for chunked encoding but we
-		 * don't handle that.  See #1320.
-		 */
-		http_Unset(bo->beresp, H_Content_Encoding);
-		bo->is_gzip = 0;
-		bo->is_gunzip = 1;
-	}
-
 	/* But we can't do both at the same time */
 	assert(bo->do_gzip == 0 || bo->do_gunzip == 0);
 
-	/* Fix Content-Encoding, as appropriate */
-	if (bo->do_gzip)
-		http_SetHeader(bo->beresp, "Content-Encoding: gzip");
-	else if (bo->do_gunzip)
-		http_Unset(bo->beresp, H_Content_Encoding);
+	if (bo->do_gunzip || (bo->is_gzip && bo->do_esi))
+		(void)VFP_Push(bo->vfc, &vfp_gunzip, 1);
 
-	if (bo->do_gunzip || (bo->is_gzip && bo->do_esi)) {
-		RFC2616_Weaken_Etag(bo->beresp);
-		http_Unset(bo->beresp, H_Content_Length);
-		VFP_Push(bo, vfp_gunzip_pull, 0);
-	}
-
-	if (bo->do_esi && bo->do_gzip) {
-		VFP_Push(bo, vfp_esi_gzip_pull, 0);
-		http_Unset(bo->beresp, H_Content_Length);
-		RFC2616_Weaken_Etag(bo->beresp);
-	} else if (bo->do_esi && bo->is_gzip && !bo->do_gunzip) {
-		VFP_Push(bo, vfp_esi_gzip_pull, 0);
-		http_Unset(bo->beresp, H_Content_Length);
-		RFC2616_Weaken_Etag(bo->beresp);
-	} else if (bo->do_esi) {
-		VFP_Push(bo, vfp_esi_pull, 0);
-	} else if (bo->do_gzip) {
-		VFP_Push(bo, vfp_gzip_pull, 0);
-		http_Unset(bo->beresp, H_Content_Length);
-		RFC2616_Weaken_Etag(bo->beresp);
-	} else if (bo->is_gzip && !bo->do_gunzip) {
-		VFP_Push(bo, vfp_testgunzip_pull, 0);
+	if (bo->htc->content_length != 0) {
+		if (bo->do_esi && bo->do_gzip) {
+			(void)VFP_Push(bo->vfc, &vfp_esi_gzip, 1);
+		} else if (bo->do_esi && bo->is_gzip && !bo->do_gunzip) {
+			(void)VFP_Push(bo->vfc, &vfp_esi_gzip, 1);
+		} else if (bo->do_esi) {
+			(void)VFP_Push(bo->vfc, &vfp_esi, 1);
+		} else if (bo->do_gzip) {
+			(void)VFP_Push(bo->vfc, &vfp_gzip, 1);
+		} else if (bo->is_gzip && !bo->do_gunzip) {
+			(void)VFP_Push(bo->vfc, &vfp_testgunzip, 1);
+		}
 	}
 
 	if (bo->fetch_objcore->flags & OC_F_PRIVATE)
 		AN(bo->uncacheable);
 
 	/* No reason to try streaming a non-existing body */
-	if (bo->htc.body_status == BS_NONE)
+	if (bo->htc->body_status == BS_NONE)
 		bo->do_stream = 0;
 
-	if (vbf_beresp2obj(bo)) {
-		(void)VFP_Error(bo, "Could not get storage");
-		VDI_CloseFd(&bo->vbc, &bo->acct);
+	if (VFP_Open(bo->vfc)) {
+		(void)VFP_Error(bo->vfc, "Fetch pipeline failed to open");
+		bo->htc->doclose = SC_RX_BODY;
+		VDI_Finish(bo->wrk, bo);
 		return (F_STP_ERROR);
 	}
 
-	assert(WRW_IsReleased(wrk));
+	if (vbf_beresp2obj(bo)) {
+		(void)VFP_Error(bo->vfc, "Could not get storage");
+		bo->htc->doclose = SC_RX_BODY;
+		VDI_Finish(bo->wrk, bo);
+		return (F_STP_ERROR);
+	}
 
-	obj = bo->fetch_obj;
+	if (bo->do_esi)
+		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_ESIPROC, 1);
 
 	if (bo->do_gzip || (bo->is_gzip && !bo->do_gunzip))
-		obj->gziped = 1;
+		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_GZIPED, 1);
 
 	if (bo->do_gzip || bo->do_gunzip)
-		obj->changed_gzip = 1;
+		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_CHGGZIP, 1);
 
-	/*
-	 * Ready to fetch the body
-	 */
+	if (http_IsStatus(bo->beresp, 200) && (
+	    http_GetHdr(bo->beresp, H_Last_Modified, &p) ||
+	    http_GetHdr(bo->beresp, H_ETag, &p)))
+		ObjSetFlag(bo->wrk, bo->fetch_objcore, OF_IMSCAND, 1);
+
+	if (bo->htc->body_status != BS_NONE)
+		AZ(VDI_GetBody(bo->wrk, bo));
 
 	assert(bo->refcount >= 1);
 
-	AZ(WS_Overflowed(bo->ws_o));
-
-	assert (bo->state == BOS_REQ_DONE);
+	assert(bo->state == BOS_REQ_DONE);
 
 	if (bo->do_stream) {
-		HSH_Unbusy(&wrk->stats, obj->objcore);
+		HSH_Unbusy(wrk, bo->fetch_objcore);
 		VBO_setstate(bo, BOS_STREAM);
 	}
 
 	VSLb(bo->vsl, SLT_Fetch_Body, "%u %s %s",
-	    bo->htc.body_status, body_status_2str(bo->htc.body_status),
+	    bo->htc->body_status, body_status_2str(bo->htc->body_status),
 	    bo->do_stream ? "stream" : "-");
 
-	if (bo->htc.body_status != BS_NONE) {
-		assert(bo->htc.body_status  != BS_ERROR);
-		VFP_Fetch_Body(bo, est);
+	if (bo->htc->body_status != BS_NONE) {
+		assert(bo->htc->body_status != BS_ERROR);
+		vbf_fetch_body_helper(bo);
 	}
 
-	if (bo->failed) {
-		AN(bo->vbc);
-		VDI_CloseFd(&bo->vbc, &bo->acct);
-	}
-
-	if (bo->failed && !bo->do_stream) {
-		assert(bo->state < BOS_STREAM);
-		if (bo->fetch_obj != NULL) {
-			oc_freeobj(bo->fetch_objcore);
-			bo->fetch_obj = NULL;
-			bo->stats->n_object--;
+	if (bo->vfc->failed) {
+		VDI_Finish(bo->wrk, bo);
+		if (!bo->do_stream) {
+			assert(bo->state < BOS_STREAM);
+			// XXX: doclose = ?
+			return (F_STP_ERROR);
+		} else {
+			return (F_STP_FAIL);
 		}
-		return (F_STP_ERROR);
 	}
 
-	if (bo->failed)
-		return (F_STP_FAIL);
-
 	if (bo->do_stream)
 		assert(bo->state == BOS_STREAM);
 	else {
 		assert(bo->state == BOS_REQ_DONE);
-		HSH_Unbusy(&wrk->stats, obj->objcore);
+		HSH_Unbusy(wrk, bo->fetch_objcore);
 	}
 
 	/* Recycle the backend connection before setting BOS_FINISHED to
 	   give predictable backend reuse behavior for varnishtest */
-	if (bo->vbc != NULL && !(bo->should_close)) {
-		VDI_RecycleFd(&bo->vbc, &bo->acct);
-		AZ(bo->vbc);
-	}
+	VDI_Finish(bo->wrk, bo);
 
 	VBO_setstate(bo, BOS_FINISHED);
 	VSLb_ts_busyobj(bo, "BerespBody", W_TIM_real(wrk));
-	if (bo->ims_obj != NULL)
-		EXP_Rearm(bo->ims_obj, bo->ims_obj->exp.t_origin, 0, 0, 0);
+	if (bo->stale_oc != NULL)
+		EXP_Rearm(bo->stale_oc, bo->stale_oc->exp.t_origin, 0, 0, 0);
 	return (F_STP_DONE);
 }
 
@@ -607,98 +696,66 @@ vbf_stp_fetch(struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_condfetch(struct worker *wrk, struct busyobj *bo)
 {
-	struct object *obj;
-	struct objiter *oi;
+	void *oi;
 	void *sp;
-	ssize_t sl, al, tl;
-	struct storage *st;
+	ssize_t sl, al, l;
+	uint8_t *ptr;
 	enum objiter_status ois;
-	char *p;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
 
-	if (bo->ims_obj->changed_gzip) {
-		/*
-		 * If we modified the gzip status of the IMS object, that
-		 * must control the C-E header, if any.
-		 */
-		http_Unset(bo->beresp, H_Content_Encoding);
-		if (http_GetHdr(bo->ims_obj->http, H_Content_Encoding, &p))
-			http_PrintfHeader(bo->beresp,
-			    "Content-Encoding: %s", p);
-	}
-
 	AZ(vbf_beresp2obj(bo));
-	obj = bo->fetch_obj;
 
-	if (bo->ims_obj->esidata != NULL) {
-		sl = bo->ims_obj->esidata->len;
-		obj->esidata = STV_alloc(bo, sl);
-		if (obj->esidata == NULL || obj->esidata->space < sl) {
-			VSLb(bo->vsl, SLT_Error,
-			    "No space for %zd bytes of ESI data", sl);
-			return (F_STP_FAIL);
-		}
-		memcpy(obj->esidata->ptr, bo->ims_obj->esidata->ptr, sl);
-		obj->esidata->len = sl;
-	}
+	if (ObjGetattr(bo->wrk, bo->stale_oc, OA_ESIDATA, NULL) != NULL)
+		AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc,
+		    OA_ESIDATA));
 
-	obj->gziped = bo->ims_obj->gziped;
-	obj->gzip_start = bo->ims_obj->gzip_start;
-	obj->gzip_last = bo->ims_obj->gzip_last;
-	obj->gzip_stop = bo->ims_obj->gzip_stop;
+	AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc, OA_FLAGS));
+	AZ(ObjCopyAttr(bo->wrk, bo->fetch_objcore, bo->stale_oc, OA_GZIPBITS));
 
-	AZ(WS_Overflowed(bo->ws_o));
 	if (bo->do_stream) {
-		HSH_Unbusy(&wrk->stats, obj->objcore);
+		HSH_Unbusy(wrk, bo->fetch_objcore);
 		VBO_setstate(bo, BOS_STREAM);
 	}
 
-	st = NULL;
 	al = 0;
-
-	oi = ObjIterBegin(wrk, bo->ims_obj);
+	oi = ObjIterBegin(wrk, bo->stale_oc);
 	do {
-		ois = ObjIter(oi, &sp, &sl);
+		ois = ObjIter(bo->stale_oc, oi, &sp, &sl);
 		if (ois == OIS_ERROR)
-			VFP_Error(bo, "Template object failed");
+			(void)VFP_Error(bo->vfc, "Template object failed");
 		while (sl > 0) {
-			if (st == NULL)
-				st = VFP_GetStorage(bo, bo->ims_obj->len - al);
-			if (st == NULL)
+			l = ObjGetLen(bo->wrk, bo->stale_oc) - al;
+			assert(l > 0);
+			if (VFP_GetStorage(bo->vfc, &l, &ptr) != VFP_OK)
 				break;
-			tl = sl;
-			if (tl > st->space - st->len)
-				tl = st->space - st->len;
-			memcpy(st->ptr + st->len, sp, tl);
-			al += tl;
-			sp = (char *)sp + tl;
-			sl -= tl;
-			VBO_extend(bo, tl);
-			if (st->len == st->space)
-				st = NULL;
+			if (sl < l)
+				l = sl;
+			memcpy(ptr, sp, l);
+			VBO_extend(bo, l);
+			al += l;
+			sp = (char *)sp + l;
+			sl -= l;
 		}
-	} while (!bo->failed && (ois == OIS_DATA || ois == OIS_STREAM));
-	ObjIterEnd(&oi);
-	if (bo->ims_obj->objcore->flags & OC_F_FAILED)
-		VFP_Error(bo, "Template object failed");
-	if (bo->failed)
+	} while (!bo->vfc->failed && (ois == OIS_DATA || ois == OIS_STREAM));
+	ObjIterEnd(bo->stale_oc, &oi);
+	if (bo->stale_oc->flags & OC_F_FAILED)
+		(void)VFP_Error(bo->vfc, "Template object failed");
+	if (bo->vfc->failed) {
+		VDI_Finish(bo->wrk, bo);
 		return (F_STP_FAIL);
+	}
 
 	if (!bo->do_stream)
-		HSH_Unbusy(&wrk->stats, obj->objcore);
+		HSH_Unbusy(wrk, bo->fetch_objcore);
 
-	assert(al == bo->ims_obj->len);
-	assert(obj->len == al);
-	EXP_Rearm(bo->ims_obj, bo->ims_obj->exp.t_origin, 0, 0, 0);
+	assert(ObjGetLen(bo->wrk, bo->fetch_objcore) == al);
+	EXP_Rearm(bo->stale_oc, bo->stale_oc->exp.t_origin, 0, 0, 0);
 
 	/* Recycle the backend connection before setting BOS_FINISHED to
 	   give predictable backend reuse behavior for varnishtest */
-	if (bo->vbc != NULL && !(bo->should_close)) {
-		VDI_RecycleFd(&bo->vbc, &bo->acct);
-		AZ(bo->vbc);
-	}
+	VDI_Finish(bo->wrk, bo);
 
 	VBO_setstate(bo, BOS_FINISHED);
 	VSLb_ts_busyobj(bo, "BerespBody", W_TIM_real(wrk));
@@ -712,73 +769,94 @@ vbf_stp_condfetch(struct worker *wrk, struct busyobj *bo)
 static enum fetch_step
 vbf_stp_error(struct worker *wrk, struct busyobj *bo)
 {
-	struct storage *st;
-	ssize_t l;
+	ssize_t l, ll, o;
 	double now;
-	char time_str[VTIM_FORMAT_SIZE];
+	uint8_t *ptr;
+	struct vsb *synth_body;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	CHECK_OBJ_NOTNULL(bo->fetch_objcore, OBJCORE_MAGIC);
+	AN(bo->fetch_objcore->flags & OC_F_BUSY);
+	assert(bo->director_state == DIR_S_NULL);
 
 	now = W_TIM_real(wrk);
 	VSLb_ts_busyobj(bo, "Error", now);
 
-	AZ(bo->vbc);		/* VDI_CloseFd must have been called before */
-	AN(bo->fetch_objcore->flags & OC_F_BUSY);
-
-	AZ(bo->synth_body);
-	bo->synth_body = VSB_new_auto();
-	AN(bo->synth_body);
+	if(bo->fetch_objcore->stobj->stevedore != NULL)
+		ObjFreeObj(bo->wrk, bo->fetch_objcore);
 
 	// XXX: reset all beresp flags ?
 
 	HTTP_Setup(bo->beresp, bo->ws, bo->vsl, SLT_BerespMethod);
 	http_PutResponse(bo->beresp, "HTTP/1.1", 503, "Backend fetch failed");
-	VTIM_format(now, time_str);
-	http_PrintfHeader(bo->beresp, "Date: %s", time_str);
+	http_TimeHeader(bo->beresp, "Date: ", now);
 	http_SetHeader(bo->beresp, "Server: Varnish");
 
-	bo->exp.t_origin = bo->t_prev;
-	bo->exp.ttl = 0;
-	bo->exp.grace = 0;
-	bo->exp.keep = 0;
+	if (bo->fetch_objcore->objhead->waitinglist != NULL) {
+		/*
+		 * If there is a waitinglist, it means that there is no
+		 * grace-able object, so cache the error return for a
+		 * short time, so the waiting list can drain, rather than
+		 * each objcore on the waiting list sequentially attempt
+		 * to fetch from the backend.
+		 */
+		bo->fetch_objcore->exp.t_origin = now;
+		bo->fetch_objcore->exp.ttl = 1;
+		bo->fetch_objcore->exp.grace = 5;
+		bo->fetch_objcore->exp.keep = 5;
+	} else {
+		EXP_Clr(&bo->fetch_objcore->exp);
+		bo->fetch_objcore->exp.t_origin = now;
+	}
 
-	VCL_backend_error_method(bo->vcl, wrk, NULL, bo, bo->bereq->ws);
+	synth_body = VSB_new_auto();
+	AN(synth_body);
 
-	AZ(VSB_finish(bo->synth_body));
+	VCL_backend_error_method(bo->vcl, wrk, NULL, bo, synth_body);
+
+	AZ(VSB_finish(synth_body));
+
+	if (wrk->handling == VCL_RET_ABANDON) {
+		VSB_delete(synth_body);
+		return (F_STP_FAIL);
+	}
 
 	if (wrk->handling == VCL_RET_RETRY) {
-		VSB_delete(bo->synth_body);
-		bo->synth_body = NULL;
-
+		VSB_delete(synth_body);
 		if (bo->retries++ < cache_param->max_retries)
 			return (F_STP_RETRY);
-
+		VSLb(bo->vsl, SLT_VCL_Error, "Too many retries, failing");
 		return (F_STP_FAIL);
 	}
 
 	assert(wrk->handling == VCL_RET_DELIVER);
 
-	if (vbf_beresp2obj(bo))
+	bo->vfc->bo = bo;
+	bo->vfc->wrk = bo->wrk;
+	bo->vfc->oc = bo->fetch_objcore;
+	bo->vfc->http = bo->beresp;
+	bo->vfc->esi_req = bo->bereq;
+
+	if (vbf_beresp2obj(bo)) {
+		VSB_delete(synth_body);
 		return (F_STP_FAIL);
-
-	l = VSB_len(bo->synth_body);
-	if (l > 0) {
-		st = VFP_GetStorage(bo, l);
-		if (st != NULL) {
-			if (st->space < l) {
-				VSLb(bo->vsl, SLT_Error,
-				    "No space for %zd bytes of synth body", l);
-			} else {
-				memcpy(st->ptr, VSB_data(bo->synth_body), l);
-				VBO_extend(bo, l);
-			}
-		}
 	}
-	VSB_delete(bo->synth_body);
-	bo->synth_body = NULL;
 
-	HSH_Unbusy(&wrk->stats, bo->fetch_objcore);
+	ll = VSB_len(synth_body);
+	o = 0;
+	while (ll > 0) {
+		l = ll;
+		if (VFP_GetStorage(bo->vfc, &l, &ptr) != VFP_OK)
+			break;
+		memcpy(ptr, VSB_data(synth_body) + o, l);
+		VBO_extend(bo, l);
+		ll -= l;
+		o += l;
+	}
+	VSB_delete(synth_body);
+
+	HSH_Unbusy(wrk, bo->fetch_objcore);
 	VBO_setstate(bo, BOS_FINISHED);
 	return (F_STP_DONE);
 }
@@ -797,10 +875,11 @@ vbf_stp_fail(struct worker *wrk, struct busyobj *bo)
 	HSH_Fail(bo->fetch_objcore);
 	if (bo->fetch_objcore->exp_flags & OC_EF_EXP) {
 		/* Already unbusied - expire it */
-		AN(bo->fetch_obj);
-		EXP_Rearm(bo->fetch_obj, bo->fetch_obj->exp.t_origin, 0, 0, 0);
+		AN(bo->fetch_objcore);
+		EXP_Rearm(bo->fetch_objcore,
+		    bo->fetch_objcore->exp.t_origin, 0, 0, 0);
 	}
-	wrk->stats.fetch_failed++;
+	wrk->stats->fetch_failed++;
 	VBO_setstate(bo, BOS_FAILED);
 	return (F_STP_DONE);
 }
@@ -812,10 +891,10 @@ static enum fetch_step
 vbf_stp_done(void)
 {
 	WRONG("Just plain wrong");
-	return (F_STP_DONE);
+	NEEDLESS_RETURN(F_STP_DONE);
 }
 
-static void
+static void __match_proto__(task_func_t)
 vbf_fetch_thread(struct worker *wrk, void *priv)
 {
 	struct busyobj *bo;
@@ -832,7 +911,17 @@ vbf_fetch_thread(struct worker *wrk, void *priv)
 	assert(isnan(bo->t_prev));
 	VSLb_ts_busyobj(bo, "Start", W_TIM_real(wrk));
 
-	bo->stats = &wrk->stats;
+	bo->wrk = wrk;
+	wrk->vsl = bo->vsl;
+
+#if 0
+	if (bo->stale_oc != NULL) {
+		CHECK_OBJ_NOTNULL(bo->stale_oc, OBJCORE_MAGIC);
+		/* We don't want the oc/stevedore ops in fetching thread */
+		if (!ObjCheckFlag(wrk, bo->stale_oc, OF_IMSCAND))
+			(void)HSH_DerefObjCore(wrk, &bo->stale_oc);
+	}
+#endif
 
 	while (stp != F_STP_DONE) {
 		CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
@@ -848,46 +937,24 @@ vbf_fetch_thread(struct worker *wrk, void *priv)
 			WRONG("Illegal fetch_step");
 		}
 	}
-	assert(WRW_IsReleased(wrk));
 
-	bo->stats = NULL;
-
-	if (bo->vbc != NULL) {
-		if (bo->should_close)
-			VDI_CloseFd(&bo->vbc, &bo->acct);
-		else
-			VDI_RecycleFd(&bo->vbc, &bo->acct);
-		AZ(bo->vbc);
-	}
+	assert(bo->director_state == DIR_S_NULL);
 
 	http_Teardown(bo->bereq);
 	http_Teardown(bo->beresp);
 
 	if (bo->state == BOS_FINISHED) {
 		AZ(bo->fetch_objcore->flags & OC_F_FAILED);
-		VSLb(bo->vsl, SLT_Length, "%zd", bo->fetch_obj->len);
-		{
-		/* Sanity check fetch methods accounting */
-			ssize_t uu;
-			struct storage *st;
-
-			uu = 0;
-			VTAILQ_FOREACH(st, &bo->fetch_obj->store, list)
-				uu += st->len;
-			if (bo->do_stream)
-				/* Streaming might have started freeing stuff */
-				assert(uu <= bo->fetch_obj->len);
-
-			else
-				assert(uu == bo->fetch_obj->len);
-		}
 		HSH_Complete(bo->fetch_objcore);
+		VSLb(bo->vsl, SLT_Length, "%ju",
+		    (uintmax_t)ObjGetLen(bo->wrk, bo->fetch_objcore));
 	}
 	AZ(bo->fetch_objcore->busyobj);
 
-	if (bo->ims_obj != NULL)
-		(void)HSH_DerefObj(&wrk->stats, &bo->ims_obj);
+	if (bo->stale_oc != NULL)
+		(void)HSH_DerefObjCore(wrk, &bo->stale_oc);
 
+	wrk->vsl = NULL;
 	VBO_DerefBusyObj(wrk, &bo);
 	THR_SetBusyobj(NULL);
 }
@@ -897,7 +964,7 @@ vbf_fetch_thread(struct worker *wrk, void *priv)
 
 void
 VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
-    struct object *oldobj, enum vbf_fetch_mode_e mode)
+    struct objcore *oldoc, enum vbf_fetch_mode_e mode)
 {
 	struct busyobj *bo, *bo_fetch;
 	const char *how;
@@ -905,11 +972,9 @@ VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-	CHECK_OBJ_ORNULL(oldobj, OBJECT_MAGIC);
+	AN(oc->flags & OC_F_BUSY);
+	CHECK_OBJ_ORNULL(oldoc, OBJCORE_MAGIC);
 
-	bo = VBO_GetBusyObj(wrk, req);
-	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
-	THR_SetBusyobj(bo);
 
 	switch(mode) {
 	case VBF_PASS:		how = "pass"; break;
@@ -918,17 +983,19 @@ VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
 	default:		WRONG("Wrong fetch mode");
 	}
 
-	VSLb(bo->vsl, SLT_Begin, "bereq %u %s",
-	    req->vsl->wid & VSL_IDENTMASK, how);
-	VSLb(req->vsl, SLT_Link, "bereq %u %s",
-	    bo->vsl->wid & VSL_IDENTMASK, how);
+	bo = VBO_GetBusyObj(wrk, req);
+	CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+	VSLb(bo->vsl, SLT_Begin, "bereq %u %s", VXID(req->vsl->wid), how);
+	VSLb(req->vsl, SLT_Link, "bereq %u %s", VXID(bo->vsl->wid), how);
+
+	THR_SetBusyobj(bo);
 
 	bo_fetch = bo;
 	bo->refcount = 2;
 
 	oc->busyobj = bo;
 
-	CHECK_OBJ_NOTNULL(bo->vcl, VCL_CONF_MAGIC);
+	AN(bo->vcl);
 
 	if (mode == VBF_PASS)
 		bo->do_pass = 1;
@@ -938,13 +1005,14 @@ VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
 
 	if (mode != VBF_BACKGROUND)
 		HSH_Ref(oc);
+	AZ(bo->fetch_objcore);
 	bo->fetch_objcore = oc;
 
-	AZ(bo->ims_obj);
-	if (oldobj != NULL) {
-		assert(oldobj->objcore->refcnt > 0);
-		HSH_Ref(oldobj->objcore);
-		bo->ims_obj = oldobj;
+	AZ(bo->stale_oc);
+	if (oldoc != NULL) {
+		assert(oldoc->refcnt > 0);
+		HSH_Ref(oldoc);
+		bo->stale_oc = oldoc;
 	}
 
 	AZ(bo->req);
@@ -953,11 +1021,11 @@ VBF_Fetch(struct worker *wrk, struct req *req, struct objcore *oc,
 	bo->fetch_task.priv = bo_fetch;
 	bo->fetch_task.func = vbf_fetch_thread;
 
-	if (Pool_Task(wrk->pool, &bo->fetch_task, POOL_QUEUE_FRONT)) {
-		wrk->stats.fetch_no_thread++;
+	if (Pool_Task(wrk->pool, &bo->fetch_task, TASK_QUEUE_BO)) {
+		wrk->stats->fetch_no_thread++;
 		(void)vbf_stp_fail(req->wrk, bo);
-		if (bo->ims_obj != NULL)
-			(void)HSH_DerefObj(&wrk->stats, &bo->ims_obj);
+		if (bo->stale_oc != NULL)
+			(void)HSH_DerefObjCore(wrk, &bo->stale_oc);
 		VBO_DerefBusyObj(wrk, &bo_fetch);
 	} else {
 		bo_fetch = NULL; /* ref transferred to fetch thread */
