@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -52,7 +52,6 @@
 
 #include "config.h"
 
-#include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
 
@@ -68,15 +67,15 @@ static struct objhead *private_oh;
 
 /*---------------------------------------------------------------------*/
 
-struct objcore *
-HSH_NewObjCore(struct worker *wrk)
+static struct objcore *
+hsh_NewObjCore(struct worker *wrk)
 {
 	struct objcore *oc;
 
 	ALLOC_OBJ(oc, OBJCORE_MAGIC);
 	XXXAN(oc);
-	wrk->stats.n_objectcore++;
-	oc->flags |= OC_F_BUSY;
+	wrk->stats->n_objectcore++;
+	oc->flags |= OC_F_BUSY | OC_F_INCOMPLETE;
 	return (oc);
 }
 
@@ -105,12 +104,12 @@ hsh_prealloc(struct worker *wrk)
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 
 	if (wrk->nobjcore == NULL)
-		wrk->nobjcore = HSH_NewObjCore(wrk);
+		wrk->nobjcore = hsh_NewObjCore(wrk);
 	CHECK_OBJ_NOTNULL(wrk->nobjcore, OBJCORE_MAGIC);
 
 	if (wrk->nobjhead == NULL) {
 		wrk->nobjhead = hsh_newobjhead();
-		wrk->stats.n_objecthead++;
+		wrk->stats->n_objecthead++;
 	}
 	CHECK_OBJ_NOTNULL(wrk->nobjhead, OBJHEAD_MAGIC);
 
@@ -119,7 +118,7 @@ hsh_prealloc(struct worker *wrk)
 		XXXAN(wl);
 		VTAILQ_INIT(&wl->list);
 		wrk->nwaitinglist = wl;
-		wrk->stats.n_waitinglist++;
+		wrk->stats->n_waitinglist++;
 	}
 	CHECK_OBJ_NOTNULL(wrk->nwaitinglist, WAITINGLIST_MAGIC);
 
@@ -136,11 +135,11 @@ HSH_Private(struct worker *wrk)
 
 	CHECK_OBJ_NOTNULL(private_oh, OBJHEAD_MAGIC);
 
-	oc = HSH_NewObjCore(wrk);
+	oc = hsh_NewObjCore(wrk);
 	AN(oc);
 	oc->refcnt = 1;
 	oc->objhead = private_oh;
-	oc->flags |= OC_F_PRIVATE;
+	oc->flags |= OC_F_PRIVATE | OC_F_BUSY;
 	Lck_Lock(&private_oh->mtx);
 	VTAILQ_INSERT_TAIL(&private_oh->objcs, oc, list);
 	private_oh->refcnt++;
@@ -155,19 +154,19 @@ HSH_Cleanup(struct worker *wrk)
 
 	if (wrk->nobjcore != NULL) {
 		FREE_OBJ(wrk->nobjcore);
-		wrk->stats.n_objectcore--;
+		wrk->stats->n_objectcore--;
 		wrk->nobjcore = NULL;
 	}
 	if (wrk->nobjhead != NULL) {
 		Lck_Delete(&wrk->nobjhead->mtx);
 		FREE_OBJ(wrk->nobjhead);
 		wrk->nobjhead = NULL;
-		wrk->stats.n_objecthead--;
+		wrk->stats->n_objecthead--;
 	}
 	if (wrk->nwaitinglist != NULL) {
 		FREE_OBJ(wrk->nwaitinglist);
 		wrk->nwaitinglist = NULL;
-		wrk->stats.n_waitinglist--;
+		wrk->stats->n_waitinglist--;
 	}
 	if (wrk->nhashpriv != NULL) {
 		/* XXX: If needed, add slinger method for this */
@@ -177,26 +176,27 @@ HSH_Cleanup(struct worker *wrk)
 }
 
 void
-HSH_DeleteObjHead(struct dstat *ds, struct objhead *oh)
+HSH_DeleteObjHead(struct worker *wrk, struct objhead *oh)
 {
 
 	AZ(oh->refcnt);
 	assert(VTAILQ_EMPTY(&oh->objcs));
 	Lck_Delete(&oh->mtx);
-	ds->n_objecthead--;
+	wrk->stats->n_objecthead--;
 	FREE_OBJ(oh);
 }
 
 void
-HSH_AddString(const struct req *req, const char *str)
+HSH_AddString(struct req *req, void *ctx, const char *str)
 {
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	AN(req->sha256ctx);
-	if (str != NULL)
-		SHA256_Update(req->sha256ctx, str, strlen(str));
-	else
-		SHA256_Update(req->sha256ctx, &str, sizeof str);
+	AN(ctx);
+	if (str != NULL) {
+		SHA256_Update(ctx, str, strlen(str));
+		VSLb(req->vsl, SLT_Hash, "%s", str);
+	} else
+		SHA256_Update(ctx, &str, 1);
 }
 
 /*---------------------------------------------------------------------
@@ -306,8 +306,8 @@ HSH_Insert(struct worker *wrk, const void *digest, struct objcore *oc)
 	/* NB: do not deref objhead the new object inherits our reference */
 	oc->objhead = oh;
 	Lck_Unlock(&oh->mtx);
-	wrk->stats.n_objectcore++;
-	wrk->stats.n_vampireobject++;
+	wrk->stats->n_objectcore++;
+	wrk->stats->n_vampireobject++;
 }
 
 /*---------------------------------------------------------------------
@@ -343,10 +343,10 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 	struct objhead *oh;
 	struct objcore *oc;
 	struct objcore *exp_oc;
-	struct object *o, *exp_o;
 	double exp_t_origin;
 	int busy_found;
 	enum lookup_e retval;
+	uint8_t *vary;
 
 	AN(ocp);
 	*ocp = NULL;
@@ -391,10 +391,8 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 
 	assert(oh->refcnt > 0);
 	busy_found = 0;
-	exp_o = NULL;
 	exp_oc = NULL;
 	exp_t_origin = 0.0;
-	o = NULL;
 	VTAILQ_FOREACH(oc, &oh->objcs, list) {
 		/* Must be at least our own ref + the objcore we examine */
 		assert(oh->refcnt > 1);
@@ -420,39 +418,37 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 			continue;
 		}
 
-		o = oc_getobj(&wrk->stats, oc);
-		CHECK_OBJ_NOTNULL(o, OBJECT_MAGIC);
-
-		if (o->exp.ttl <= 0.)
-			continue;
-		if (BAN_CheckObject(o, req))
-			continue;
-		if (o->vary != NULL && !VRY_Match(req, o->vary))
+		if (oc->exp.ttl <= 0.)
 			continue;
 
-		if (EXP_Ttl(req, o) >= req->t_req) {
+		if (BAN_CheckObject(wrk, oc, req))
+			continue;
+
+		vary = ObjGetattr(wrk, oc, OA_VARY, NULL);
+
+		if (vary != NULL && !VRY_Match(req, vary))
+			continue;
+
+		if (EXP_Ttl(req, &oc->exp) >= req->t_req) {
 			/* If still valid, use it */
 			assert(oh->refcnt > 1);
 			assert(oc->objhead == oh);
 			oc->refcnt++;
-			if (oh->hits < LONG_MAX)
-				oh->hits++;
+			if (oc->hits < LONG_MAX)
+				oc->hits++;
 			Lck_Unlock(&oh->mtx);
-			assert(HSH_DerefObjHead(&wrk->stats, &oh));
+			assert(HSH_DerefObjHead(wrk, &oh));
 			*ocp = oc;
 			return (HSH_HIT);
 		}
-		if (o->exp.t_origin > exp_t_origin &&
-		    !(oc->flags & OC_F_PASS)) {
+		if (oc->exp.t_origin > exp_t_origin) {
 			/* record the newest object */
 			exp_oc = oc;
-			exp_o = o;
-			exp_t_origin = o->exp.t_origin;
+			exp_t_origin = oc->exp.t_origin;
 		}
 	}
 
 	if (exp_oc != NULL) {
-		AN(exp_o);
 		assert(oh->refcnt > 1);
 		assert(exp_oc->objhead == oh);
 		exp_oc->refcnt++;
@@ -464,11 +460,11 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 			AZ(req->hash_ignore_busy);
 			retval = HSH_EXP;
 		}
-		if (oh->hits < LONG_MAX)
-			oh->hits++;
+		if (exp_oc->hits < LONG_MAX)
+			exp_oc->hits++;
 		Lck_Unlock(&oh->mtx);
 		if (retval == HSH_EXP)
-			assert(HSH_DerefObjHead(&wrk->stats, &oh));
+			assert(HSH_DerefObjHead(wrk, &oh));
 		*ocp = exp_oc;
 		return (retval);
 	}
@@ -500,7 +496,7 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
 			VSLb(req->vsl, SLT_Debug, "hit busy obj <%p>", oh);
 	}
 
-	wrk->stats.busy_sleep++;
+	wrk->stats->busy_sleep++;
 	/*
 	 * The objhead reference transfers to the sess, we get it
 	 * back when the sess comes off the waiting list and
@@ -516,37 +512,59 @@ HSH_Lookup(struct req *req, struct objcore **ocp, struct objcore **bocp,
  */
 
 static void
-hsh_rush(struct dstat *ds, struct objhead *oh)
+hsh_rush(struct worker *wrk, struct objhead *oh)
 {
 	unsigned u;
 	struct req *req;
+	struct sess *sp;
 	struct waitinglist *wl;
 
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 	Lck_AssertHeld(&oh->mtx);
 	wl = oh->waitinglist;
-	CHECK_OBJ_NOTNULL(wl, WAITINGLIST_MAGIC);
+	CHECK_OBJ_ORNULL(wl, WAITINGLIST_MAGIC);
+	if (wl == NULL)
+		return;
 	for (u = 0; u < cache_param->rush_exponent; u++) {
 		req = VTAILQ_FIRST(&wl->list);
 		if (req == NULL)
 			break;
 		CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-		ds->busy_wakeup++;
+		wrk->stats->busy_wakeup++;
 		AZ(req->wrk);
 		VTAILQ_REMOVE(&wl->list, req, w_list);
 		DSL(DBG_WAITINGLIST, req->vsl->wid, "off waiting list");
-		if (SES_ScheduleReq(req)) {
+		if (SES_Reschedule_Req(req)) {
 			/*
-			 * We could not schedule the session, put it back.
+			 * In case of overloads, we ditch the entire
+			 * waiting list.
 			 */
-			VTAILQ_INSERT_HEAD(&wl->list, req, w_list);
+			wrk->stats->busy_wakeup--;
+			while (1) {
+				wrk->stats->busy_killed++;
+				AN (req->vcl);
+				VCL_Rel(&req->vcl);
+				sp = req->sp;
+				CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+				CNT_AcctLogCharge(wrk->stats, req);
+				Req_Release(req);
+				SES_Delete(sp, SC_OVERLOAD, NAN);
+				req = VTAILQ_FIRST(&wl->list);
+				if (req == NULL)
+					break;
+				CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+				VTAILQ_REMOVE(&wl->list, req, w_list);
+				DSL(DBG_WAITINGLIST, req->vsl->wid,
+				    "kill from waiting list");
+			}
 			break;
 		}
 	}
 	if (VTAILQ_EMPTY(&wl->list)) {
 		oh->waitinglist = NULL;
 		FREE_OBJ(wl);
-		ds->n_waitinglist--;
+		wrk->stats->n_waitinglist--;
 	}
 }
 
@@ -560,9 +578,7 @@ double keep)
 {
 	struct objcore *oc, **ocp;
 	unsigned spc, ospc, nobj, n;
-	struct object *o;
 	int more = 0;
-
 	double now;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
@@ -606,34 +622,14 @@ double keep)
 		for (n = 0; n < nobj; n++) {
 			oc = ocp[n];
 			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
-			o = oc_getobj(&wrk->stats, oc);
-			if (o == NULL)
-				continue;
-			CHECK_OBJ_NOTNULL(o, OBJECT_MAGIC);
-			EXP_Rearm(o, now, ttl, grace, keep);
-			(void)HSH_DerefObj(&wrk->stats, &o);
+			EXP_Rearm(oc, now, ttl, grace, keep);
+			(void)HSH_DerefObjCore(wrk, &oc);
 		}
 	} while (more);
 	WS_Release(wrk->aws, 0);
 	Pool_PurgeStat(nobj);
 }
 
-
-/*---------------------------------------------------------------------
- * Kill a busy object we don't need and can't use.
- */
-
-void
-HSH_Drop(struct worker *wrk, struct object **oo)
-{
-
-	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
-	AN(oo);
-	CHECK_OBJ_NOTNULL(*oo, OBJECT_MAGIC);
-	(*oo)->exp.ttl = -1.;
-	AZ(HSH_DerefObj(&wrk->stats, oo));
-}
-
 /*---------------------------------------------------------------------
  * Fail an objcore
  */
@@ -652,10 +648,11 @@ HSH_Fail(struct objcore *oc)
 	 * will not consider this oc, or an object hung of the oc
 	 * so that it can consider it.
 	 */
-	assert((oc->flags & OC_F_BUSY) || (oc->methods != NULL));
+	assert((oc->flags & OC_F_BUSY) || (oc->stobj->stevedore != NULL));
 
 	Lck_Lock(&oh->mtx);
 	oc->flags |= OC_F_FAILED;
+	oc->flags &= ~OC_F_INCOMPLETE;
 	oc->busyobj = NULL;
 	Lck_Unlock(&oh->mtx);
 }
@@ -675,6 +672,7 @@ HSH_Complete(struct objcore *oc)
 
 	Lck_Lock(&oh->mtx);
 	oc->busyobj = NULL;
+	oc->flags &= ~OC_F_INCOMPLETE;
 	Lck_Unlock(&oh->mtx);
 }
 
@@ -683,21 +681,22 @@ HSH_Complete(struct objcore *oc)
  */
 
 void
-HSH_Unbusy(struct dstat *ds, struct objcore *oc)
+HSH_Unbusy(struct worker *wrk, struct objcore *oc)
 {
 	struct objhead *oh;
 
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	oh = oc->objhead;
 	CHECK_OBJ(oh, OBJHEAD_MAGIC);
 
-	AN(oc->methods);
+	AN(oc->stobj->stevedore);
 	AN(oc->flags & OC_F_BUSY);
 	assert(oh->refcnt > 0);
 
 	if (!(oc->flags & OC_F_PRIVATE)) {
 		BAN_NewObjCore(oc);
-		EXP_Insert(oc);
+		EXP_Insert(wrk, oc);
 		AN(oc->exp_flags & OC_EF_EXP);
 		AN(oc->ban);
 	}
@@ -710,7 +709,7 @@ HSH_Unbusy(struct dstat *ds, struct objcore *oc)
 	VTAILQ_INSERT_HEAD(&oh->objcs, oc, list);
 	oc->flags &= ~OC_F_BUSY;
 	if (oh->waitinglist != NULL)
-		hsh_rush(ds, oh);
+		hsh_rush(wrk, oh);
 	Lck_Unlock(&oh->mtx);
 }
 
@@ -756,46 +755,23 @@ HSH_RefBusy(const struct objcore *oc)
 }
 
 /*--------------------------------------------------------------------
- * Dereference objcore and or object
- *
- * Can deal with:
- *	bare objcore (incomplete fetch)
- *	bare object (pass)
- *	object with objcore
- *	XXX later:  objcore with object (?)
- *
- * But you can only supply one of the two arguments at a time.
+ * Dereference objcore
  *
  * Returns zero if target was destroyed.
  */
 
 int
-HSH_DerefObj(struct dstat *ds, struct object **oo)
-{
-	struct object *o;
-	struct objcore *oc;
-
-	AN(oo);
-	o = *oo;
-	*oo = NULL;
-
-	CHECK_OBJ_NOTNULL(o, OBJECT_MAGIC);
-	oc = o->objcore;
-	return (HSH_DerefObjCore(ds, &oc));
-}
-
-int
-HSH_DerefObjCore(struct dstat *ds, struct objcore **ocp)
+HSH_DerefObjCore(struct worker *wrk, struct objcore **ocp)
 {
 	struct objcore *oc;
 	struct objhead *oh;
 	unsigned r;
 
 	AN(ocp);
-	AN(ds);
 	oc = *ocp;
 	*ocp = NULL;
 
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	assert(oc->refcnt > 0);
 
@@ -808,7 +784,7 @@ HSH_DerefObjCore(struct dstat *ds, struct objcore **ocp)
 	if (!r)
 		VTAILQ_REMOVE(&oh->objcs, oc, list);
 	if (oh->waitinglist != NULL)
-		hsh_rush(ds, oh);
+		hsh_rush(wrk, oh);
 	Lck_Unlock(&oh->mtx);
 	if (r != 0)
 		return (r);
@@ -816,32 +792,31 @@ HSH_DerefObjCore(struct dstat *ds, struct objcore **ocp)
 	BAN_DestroyObj(oc);
 	AZ(oc->ban);
 
-	if (oc->methods != NULL) {
-		oc_freeobj(oc);
-		ds->n_object--;
-	}
+	if (oc->stobj->stevedore != NULL)
+		ObjFreeObj(wrk, oc);
 	FREE_OBJ(oc);
 
-	ds->n_objectcore--;
+	wrk->stats->n_objectcore--;
 	/* Drop our ref on the objhead */
 	assert(oh->refcnt > 0);
-	(void)HSH_DerefObjHead(ds, &oh);
+	(void)HSH_DerefObjHead(wrk, &oh);
 	return (0);
 }
 
 int
-HSH_DerefObjHead(struct dstat *ds, struct objhead **poh)
+HSH_DerefObjHead(struct worker *wrk, struct objhead **poh)
 {
 	struct objhead *oh;
 	int r;
 
-	AN(ds);
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(poh);
 	oh = *poh;
 	*poh = NULL;
 	CHECK_OBJ_NOTNULL(oh, OBJHEAD_MAGIC);
 
 	if (oh == private_oh) {
+		AZ(oh->waitinglist);
 		Lck_Lock(&oh->mtx);
 		assert(oh->refcnt > 1);
 		oh->refcnt--;
@@ -849,10 +824,29 @@ HSH_DerefObjHead(struct dstat *ds, struct objhead **poh)
 		return(1);
 	}
 
+	/*
+	 * Make absolutely certain that we do not let the final ref
+	 * disappear until the waitinglist is empty.
+	 * This is necessary because the req's on the waiting list do
+	 * not hold any ref on the objhead of their own, and we cannot
+	 * just make the hold the same ref's as objcore, that would
+	 * confuse hashers.
+	 */
+	while (oh->waitinglist != NULL) {
+		Lck_Lock(&oh->mtx);
+		assert(oh->refcnt > 0);
+		r = oh->refcnt;
+		hsh_rush(wrk, oh);
+		Lck_Unlock(&oh->mtx);
+		if (r > 1)
+			break;
+		usleep(100000);
+	}
+
 	assert(oh->refcnt > 0);
 	r = hash->deref(oh);
 	if (!r)
-		HSH_DeleteObjHead(ds, oh);
+		HSH_DeleteObjHead(wrk, oh);
 	return (r);
 }
 
