@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -36,12 +36,18 @@
 
 #include "cache.h"
 
+#include "vend.h"
 #include "vct.h"
+#include "vtim.h"
 
 #define HTTPH(a, b, c) char b[] = "*" a ":";
 #include "tbl/http_headers.h"
 #undef HTTPH
 
+const char H__Status[]	= "\010:status:";
+const char H__Proto[]	= "\007:proto:";
+const char H__Reason[]	= "\010:reason:";
+
 /*--------------------------------------------------------------------
  * These two functions are in an incestous relationship with the
  * order of macros in include/tbl/vsl_tags_http.h
@@ -95,12 +101,12 @@ http_VSL_log(const struct http *hp)
 /*--------------------------------------------------------------------*/
 
 static void
-http_fail(struct http *hp)
+http_fail(const struct http *hp)
 {
 
 	VSC_C_main->losthdr++;
-	VSLb(hp->vsl, SLT_Error, "out of workspace");
-	hp->failed = 1;
+	VSLb(hp->vsl, SLT_Error, "out of workspace (%s)", hp->ws->id);
+	WS_MarkOverflow(hp->ws);
 }
 
 /*--------------------------------------------------------------------*/
@@ -135,7 +141,7 @@ HTTP_estimate(unsigned nhttp)
 {
 
 	/* XXX: We trust the structs to size-aligned as necessary */
-	return (sizeof (struct http) + (sizeof (txt) + 1) * nhttp);
+	return (PRNDUP(sizeof (struct http) + sizeof(txt) * nhttp + nhttp));
 }
 
 struct http *
@@ -199,7 +205,7 @@ void
 http_SetH(const struct http *to, unsigned n, const char *fm)
 {
 
-	assert(n < to->shd);
+	assert(n < to->nhd);
 	AN(fm);
 	to->hd[n].b = TRUST_ME(fm);
 	to->hd[n].e = strchr(to->hd[n].b, '\0');
@@ -210,7 +216,7 @@ http_SetH(const struct http *to, unsigned n, const char *fm)
 /*--------------------------------------------------------------------*/
 
 static void
-http_PutField(struct http *to, int field, const char *string)
+http_PutField(const struct http *to, int field, const char *string)
 {
 	char *p;
 
@@ -229,7 +235,7 @@ http_PutField(struct http *to, int field, const char *string)
 
 /*--------------------------------------------------------------------*/
 
-int
+static int
 http_IsHdr(const txt *hh, const char *hdr)
 {
 	unsigned l;
@@ -263,6 +269,26 @@ http_findhdr(const struct http *hp, unsigned l, const char *hdr)
 	return (0);
 }
 
+/*--------------------------------------------------------------------
+ * Count how many instances we have of this header
+ */
+
+unsigned
+http_CountHdr(const struct http *hp, const char *hdr)
+{
+	unsigned retval = 0;
+	unsigned u;
+
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+
+	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
+		Tcheck(hp->hd[u]);
+		if (http_IsHdr(&hp->hd[u], hdr))
+			retval++;
+	}
+	return (retval);
+}
+
 /*--------------------------------------------------------------------
  * This function collapses multiple headerlines of the same name.
  * The lines are joined with a comma, according to [rfc2616, 4.2bot, p32]
@@ -274,7 +300,8 @@ http_CollectHdr(struct http *hp, const char *hdr)
 	unsigned u, l, ml, f, x, d;
 	char *b = NULL, *e = NULL;
 
-	if (hp->failed)
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+	if (WS_Overflowed(hp->ws))
 		return;
 	l = hdr[0];
 	assert(l == strlen(hdr + 1));
@@ -338,10 +365,10 @@ http_CollectHdr(struct http *hp, const char *hdr)
 /*--------------------------------------------------------------------*/
 
 int
-http_GetHdr(const struct http *hp, const char *hdr, char **ptr)
+http_GetHdr(const struct http *hp, const char *hdr, const char **ptr)
 {
 	unsigned u, l;
-	char *p;
+	const char *p;
 
 	l = hdr[0];
 	assert(l == strlen(hdr + 1));
@@ -362,47 +389,139 @@ http_GetHdr(const struct http *hp, const char *hdr, char **ptr)
 	return (1);
 }
 
-/*--------------------------------------------------------------------
- * Find a given data element in a header according to RFC2616's #rule
+/*-----------------------------------------------------------------------------
+ * Split source string at any of the separators, return pointer to first
+ * and last+1 char of substrings, with whitespace trimed at both ends.
+ * If sep being an empty string is shorthand for VCT::SP
+ * If stop is NULL, src is NUL terminated.
+ */
+
+static int
+http_split(const char **src, const char *stop, const char *sep,
+    const char **b, const char **e)
+{
+	const char *p, *q;
+
+	AN(src);
+	AN(*src);
+	AN(sep);
+	AN(b);
+	AN(e);
+
+	if (stop == NULL)
+		stop = strchr(*src, '\0');
+
+	for (p = *src; p < stop && (vct_issp(*p) || strchr(sep, *p)); p++)
+		continue;
+
+	if (p >= stop) {
+		*b = NULL;
+		*e = NULL;
+		return (0);
+	}
+
+	*b = p;
+	if (*sep == '\0') {
+		for (q = p + 1; q < stop && !vct_issp(*q); q++)
+			continue;
+		*e = q;
+		*src = q;
+		return (1);
+	}
+	for (q = p + 1; q < stop && !strchr(sep, *q); q++)
+		continue;
+	*src = q;
+	while (q > p && vct_issp(q[-1]))
+		q--;
+	*e = q;
+	return (1);
+}
+
+/*-----------------------------------------------------------------------------
+ * Comparison rule for tokens:
+ *	if target string starts with '"', we use memcmp() and expect closing
+ *	double quote as well
+ *	otherwise we use strncascmp()
+ *
+ * On match we increment *bp past the token name.
+ */
+
+static int
+http_istoken(const char **bp, const char *e, const char *token)
+{
+	int fl = strlen(token);
+	const char *b;
+
+	AN(bp);
+	AN(e);
+	AN(token);
+
+	b = *bp;
+
+	if (b + fl + 2 <= e && *b == '"' &&
+	    !memcmp(b + 1, token, fl) && b[fl + 1] == '"') {
+		*bp += fl + 2;
+		return (1);
+	}
+	if (b + fl <= e && !strncasecmp(b, token, fl) &&
+	    (b + fl == e || !vct_istchar(b[fl]))) {
+		*bp += fl;
+		return (1);
+	}
+	return (0);
+}
+
+/*-----------------------------------------------------------------------------
+ * Find a given data element (token) in a header according to RFC2616's #rule
  * (section 2.1, p15)
+ *
+ * On case sensitivity:
+ *
+ * Section 4.2 (Messages Headers) defines field (header) name as case
+ * insensitive, but the field (header) value/content may be case-sensitive.
+ *
+ * http_GetHdrToken looks up a token in a header value and the rfc does not say
+ * explicitly if tokens are to be compared with or without respect to case.
+ *
+ * But all examples and specific statements regarding tokens follow the rule
+ * that unquoted tokens are to be matched case-insensitively and quoted tokens
+ * case-sensitively.
+ *
+ * The optional pb and pe arguments will point to the token content start and
+ * end+1, white space trimmed on both sides.
  */
 
 int
-http_GetHdrData(const struct http *hp, const char *hdr,
-    const char *field, char **ptr)
+http_GetHdrToken(const struct http *hp, const char *hdr,
+    const char *token, const char **pb, const char **pe)
 {
-	char *h, *e;
-	unsigned fl;
+	const char *h, *b, *e;
 
-	if (ptr != NULL)
-		*ptr = NULL;
+	if (pb != NULL)
+		*pb = NULL;
+	if (pe != NULL)
+		*pe = NULL;
 	if (!http_GetHdr(hp, hdr, &h))
 		return (0);
 	AN(h);
-	e = strchr(h, '\0');
-	fl = strlen(field);
-	while (h + fl <= e) {
-		/* Skip leading whitespace and commas */
-		if (vct_islws(*h) || *h == ',') {
-			h++;
+
+	while(http_split(&h, NULL, ",", &b, &e))
+		if (http_istoken(&b, e, token))
+			break;
+	if (b == NULL)
+		return (0);
+	if (pb != NULL) {
+		for (; vct_islws(*b); b++)
 			continue;
+		if (b == e) {
+			b = NULL;
+			e = NULL;
 		}
-		/* Check for substrings before memcmp() */
-		if ((h + fl == e || vct_issepctl(h[fl])) &&
-		    !memcmp(h, field, fl)) {
-			if (ptr != NULL) {
-				h += fl;
-				while (vct_islws(*h))
-					h++;
-				*ptr = h;
-			}
-			return (1);
-		}
-		/* Skip until end of header or comma */
-		while (*h && *h != ',')
-			h++;
+		*pb = b;
+		if (pe != NULL)
+			*pe = e;
 	}
-	return (0);
+	return (1);
 }
 
 /*--------------------------------------------------------------------
@@ -412,45 +531,44 @@ http_GetHdrData(const struct http *hp, const char *hdr,
 double
 http_GetHdrQ(const struct http *hp, const char *hdr, const char *field)
 {
-	char *h;
+	const char *hb, *he, *b, *e;
 	int i;
-	double a, b;
+	double a, f;
 
-	h = NULL;
-	i = http_GetHdrData(hp, hdr, field, &h);
+	i = http_GetHdrToken(hp, hdr, field, &hb, &he);
 	if (!i)
 		return (0.);
 
-	if (h == NULL)
+	if (hb == NULL)
 		return (1.);
-	/* Skip whitespace, looking for '=' */
-	while (*h && vct_issp(*h))
-		h++;
-	if (*h++ != ';')
-		return (1.);
-	while (*h && vct_issp(*h))
-		h++;
-	if (*h++ != 'q')
-		return (1.);
-	while (*h && vct_issp(*h))
-		h++;
-	if (*h++ != '=')
-		return (1.);
-	while (*h && vct_issp(*h))
-		h++;
-	a = 0.;
-	while (vct_isdigit(*h)) {
-		a *= 10.;
-		a += *h - '0';
-		h++;
+	while(http_split(&hb, he, ";", &b, &e)) {
+		if (*b != 'q')
+			continue;
+		for (b++; b < e && vct_issp(*b); b++)
+			continue;
+		if (b == e || *b != '=')
+			continue;
+		break;
 	}
-	if (*h++ != '.')
+	if (b == NULL)
+		return (1.);
+	for (b++; b < e && vct_issp(*b); b++)
+		continue;
+	if (b == e || (*b != '.' && !vct_isdigit(*b)))
+		return (0.);
+	a = 0;
+	while (b < e && vct_isdigit(*b)) {
+		a *= 10.;
+		a += *b - '0';
+		b++;
+	}
+	if (b == e || *b++ != '.')
 		return (a);
-	b = .1;
-	while (vct_isdigit(*h)) {
-		a += b * (*h - '0');
-		b *= .1;
-		h++;
+	f = .1;
+	while (b < e && vct_isdigit(*b)) {
+		a += f * (*b - '0');
+		f *= .1;
+		b++;
 	}
 	return (a);
 }
@@ -461,16 +579,16 @@ http_GetHdrQ(const struct http *hp, const char *hdr, const char *field)
 
 int
 http_GetHdrField(const struct http *hp, const char *hdr,
-    const char *field, char **ptr)
+    const char *field, const char **ptr)
 {
-	char *h;
+	const char *h;
 	int i;
 
 	if (ptr != NULL)
 		*ptr = NULL;
 
 	h = NULL;
-	i = http_GetHdrData(hp, hdr, field, &h);
+	i = http_GetHdrToken(hp, hdr, field, &h, NULL);
 	if (!i)
 		return (i);
 
@@ -488,43 +606,76 @@ http_GetHdrField(const struct http *hp, const char *hdr,
 	return (i);
 }
 
+/*--------------------------------------------------------------------*/
+
+ssize_t
+http_GetContentLength(const struct http *hp)
+{
+	ssize_t cl, cll;
+	const char *b;
+
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+
+	if (!http_GetHdr(hp, H_Content_Length, &b))
+		return (-1);
+	cl = 0;
+	if (!vct_isdigit(*b))
+		return (-2);
+	for (;vct_isdigit(*b); b++) {
+		cll = cl;
+		cl *= 10;
+		cl += *b - '0';
+		if (cll != cl / 10)
+			return (-2);
+	}
+	while (vct_islws(*b))
+		b++;
+	if (*b != '\0')
+		return (-2);
+	return (cl);
+}
+
 /*--------------------------------------------------------------------
- * XXX: redo with http_GetHdrField() ?
  */
 
 enum sess_close
-http_DoConnection(const struct http *hp)
+http_DoConnection(struct http *hp)
 {
-	char *p, *q;
-	enum sess_close ret;
-	unsigned u;
+	const char *h, *b, *e;
+	enum sess_close retval;
+	unsigned u, v;
 
-	if (!http_GetHdr(hp, H_Connection, &p)) {
-		if (hp->protover < 11)
-			return (SC_REQ_HTTP10);
-		return (SC_NULL);
+	if (hp->protover == 10)
+		retval = SC_REQ_HTTP10;
+	else
+		retval = SC_NULL;
+
+	http_CollectHdr(hp, H_Connection);
+	if (!http_GetHdr(hp, H_Connection, &h))
+		return (retval);
+	AN(h);
+	while (http_split(&h, NULL, ",", &b, &e)) {
+		u = pdiff(b, e);
+		if (u == 5 && !strncasecmp(b, "close", u))
+			retval = SC_REQ_CLOSE;
+		if (u == 10 && !strncasecmp(b, "keep-alive", u))
+			retval = SC_NULL;
+
+		/* Refuse removal of well-known-headers if they would pass. */
+/*lint -save -e506 */
+#define HTTPH(a, x, c)						\
+		if (!((c) & HTTPH_R_PASS) &&			\
+		    strlen(a) == u && !strncasecmp(a, b, u))	\
+			return (SC_RX_BAD);
+#include "tbl/http_headers.h"
+#undef HTTPH
+/*lint -restore */
+
+		v = http_findhdr(hp, u, b);
+		if (v > 0)
+			hp->hdf[v] |= HDF_FILTER;
 	}
-	ret = SC_NULL;
-	AN(p);
-	for (; *p; p++) {
-		if (vct_issp(*p))
-			continue;
-		if (*p == ',')
-			continue;
-		for (q = p + 1; *q; q++)
-			if (*q == ',' || vct_issp(*q))
-				break;
-		u = pdiff(p, q);
-		if (u == 5 && !strncasecmp(p, "close", u))
-			ret = SC_REQ_CLOSE;
-		u = http_findhdr(hp, u, p);
-		if (u != 0)
-			hp->hdf[u] |= HDF_FILTER;
-		if (!*q)
-			break;
-		p = q;
-	}
-	return (ret);
+	return (retval);
 }
 
 /*--------------------------------------------------------------------*/
@@ -532,7 +683,7 @@ http_DoConnection(const struct http *hp)
 int
 http_HdrIs(const struct http *hp, const char *hdr, const char *val)
 {
-	char *p;
+	const char *p;
 
 	if (!http_GetHdr(hp, hdr, &p))
 		return (0);
@@ -552,6 +703,15 @@ http_GetStatus(const struct http *hp)
 	return (hp->status);
 }
 
+int
+http_IsStatus(const struct http *hp, int val)
+{
+
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+	assert(val >= 100 && val <= 999);
+	return (val == (hp->status % 1000));
+}
+
 /*--------------------------------------------------------------------
  * Setting the status will also set the Reason appropriately
  */
@@ -577,7 +737,7 @@ http_SetStatus(struct http *to, uint16_t status)
 /*--------------------------------------------------------------------*/
 
 const char *
-http_GetReq(const struct http *hp)
+http_GetMethod(const struct http *hp)
 {
 
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
@@ -621,16 +781,17 @@ http_PutResponse(struct http *to, const char *proto, uint16_t status,
  */
 
 unsigned
-http_EstimateWS(const struct http *fm, unsigned how, uint16_t *nhd)
+http_EstimateWS(const struct http *fm, unsigned how)
 {
 	unsigned u, l;
 
-	l = 0;
-	*nhd = HTTP_HDR_FIRST;
+	l = 4;
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
 	for (u = 0; u < fm->nhd; u++) {
-		if (fm->hd[u].b == NULL)
+		if (u == HTTP_HDR_METHOD || u == HTTP_HDR_URL)
 			continue;
+		AN(fm->hd[u].b);
+		AN(fm->hd[u].e);
 		if (fm->hdf[u] & HDF_FILTER)
 			continue;
 #define HTTPH(a, b, c) \
@@ -638,11 +799,217 @@ http_EstimateWS(const struct http *fm, unsigned how, uint16_t *nhd)
 			continue;
 #include "tbl/http_headers.h"
 #undef HTTPH
-		l += PRNDUP(Tlen(fm->hd[u]) + 1);
-		(*nhd)++;
-		// fm->hdf[u] |= HDF_COPY;
+		l += Tlen(fm->hd[u]) + 1L;
+	}
+	return (PRNDUP(l + 1L));
+}
+
+/*--------------------------------------------------------------------
+ * Encode http struct as byte string.
+ *
+ * XXX: We could save considerable special-casing below by encoding also
+ * XXX: H__Status, H__Reason and H__Proto into the string, but it would
+ * XXX: add 26-30 bytes to all encoded objects to save a little code.
+ * XXX: It could possibly be a good idea for later HTTP versions.
+ */
+
+void
+HTTP_Encode(const struct http *fm, uint8_t *p0, unsigned l, unsigned how)
+{
+	unsigned u, w;
+	uint16_t n;
+	uint8_t *p, *e;
+
+	AN(p0);
+	AN(l);
+	p = p0;
+	e = p + l;
+	assert(p + 5 <= e);
+	assert(fm->nhd <= fm->shd);
+	n = HTTP_HDR_FIRST - 3;
+	vbe16enc(p + 2, fm->status);
+	p += 4;
+	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
+	for (u = 0; u < fm->nhd; u++) {
+		if (u == HTTP_HDR_METHOD || u == HTTP_HDR_URL)
+			continue;
+		AN(fm->hd[u].b);
+		AN(fm->hd[u].e);
+		if (fm->hdf[u] & HDF_FILTER)
+			continue;
+#define HTTPH(a, b, c) \
+		if (((c) & how) && http_IsHdr(&fm->hd[u], (b))) \
+			continue;
+#include "tbl/http_headers.h"
+#undef HTTPH
+		http_VSLH(fm, u);
+		w = Tlen(fm->hd[u]) + 1L;
+		assert(p + w + 1 <= e);
+		memcpy(p, fm->hd[u].b, w);
+		p += w;
+		n++;
+	}
+	*p++ = '\0';
+	assert(p <= e);
+	vbe16enc(p0, n + 1);
+}
+
+/*--------------------------------------------------------------------
+ * Decode byte string into http struct
+ */
+
+int
+HTTP_Decode(struct http *to, const uint8_t *fm)
+{
+
+	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
+	AN(to->vsl);
+	AN(fm);
+	if (vbe16dec(fm) <= to->shd) {
+		to->status = vbe16dec(fm + 2);
+		fm += 4;
+		for (to->nhd = 0; to->nhd < to->shd; to->nhd++) {
+			if (to->nhd == HTTP_HDR_METHOD ||
+			    to->nhd == HTTP_HDR_URL) {
+				to->hd[to->nhd].b = NULL;
+				to->hd[to->nhd].e = NULL;
+				continue;
+			}
+			if (*fm == '\0')
+				return (0);
+			to->hd[to->nhd].b = (const void*)fm;
+			fm = (const void*)strchr((const void*)fm, '\0');
+			to->hd[to->nhd].e = (const void*)fm;
+			fm++;
+			http_VSLH(to, to->nhd);
+		}
+	}
+	VSLb(to->vsl, SLT_Error,
+	    "Too many headers to Decode object (%u vs. %u)",
+	    vbe16dec(fm), to->shd);
+	return (-1);
+}
+
+/*--------------------------------------------------------------------*/
+
+uint16_t
+HTTP_GetStatusPack(struct worker *wrk, struct objcore *oc)
+{
+	const char *ptr;
+	ptr = ObjGetattr(wrk, oc, OA_HEADERS, NULL);
+	AN(ptr);
+
+	return(vbe16dec(ptr + 2));
+}
+
+/*--------------------------------------------------------------------*/
+
+/* Get the first packed header */
+int
+HTTP_IterHdrPack(struct worker *wrk, struct objcore *oc, const char **p)
+{
+	const char *ptr;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	AN(p);
+
+	if (*p == NULL) {
+		ptr = ObjGetattr(wrk, oc, OA_HEADERS, NULL);
+		AN(ptr);
+		ptr += 4;	/* Skip nhd and status */
+		ptr = strchr(ptr, '\0') + 1;	/* Skip :proto: */
+		ptr = strchr(ptr, '\0') + 1;	/* Skip :status: */
+		ptr = strchr(ptr, '\0') + 1;	/* Skip :reason: */
+		*p = ptr;
+	} else {
+		*p = strchr(*p, '\0') + 1;	/* Skip to next header */
+	}
+	if (**p == '\0')
+		return (0);
+	return (1);
+}
+
+const char *
+HTTP_GetHdrPack(struct worker *wrk, struct objcore *oc, const char *hdr)
+{
+	const char *ptr;
+	unsigned l;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	AN(hdr);
+
+	l = hdr[0];
+	assert(l > 0);
+	assert(l == strlen(hdr + 1));
+	assert(hdr[l] == ':');
+	hdr++;
+
+	if (hdr[0] == ':') {
+		/* Special cases */
+		ptr = ObjGetattr(wrk, oc, OA_HEADERS, NULL);
+		AN(ptr);
+		ptr += 4;	/* Skip nhd and status */
+
+		if (!strcmp(hdr, ":proto:"))
+			return (ptr);
+		ptr = strchr(ptr, '\0') + 1;
+		if (!strcmp(hdr, ":status:"))
+			return (ptr);
+		ptr = strchr(ptr, '\0') + 1;
+		if (!strcmp(hdr, ":reason:"))
+			return (ptr);
+		WRONG("Unknown magic packed header");
+	}
+
+	HTTP_FOREACH_PACK(wrk, oc, ptr) {
+		if (!strncasecmp(ptr, hdr, l)) {
+			ptr += l;
+			while (vct_islws(*ptr))
+				ptr++;
+			return (ptr);
+		}
+	}
+
+	return (NULL);
+}
+
+/*--------------------------------------------------------------------
+ * Merge any headers in the oc->OA_HEADER into the struct http if they
+ * are not there already.
+ */
+
+void
+HTTP_Merge(struct worker *wrk, struct objcore *oc, struct http *to)
+{
+	const char *ptr;
+	unsigned u;
+	const char *p;
+
+	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
+	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
+
+	ptr = ObjGetattr(wrk, oc, OA_HEADERS, NULL);
+	AN(ptr);
+
+	to->status = vbe16dec(ptr + 2);
+	ptr += 4;
+
+	for (u = 0; u < HTTP_HDR_FIRST; u++) {
+		if (u == HTTP_HDR_METHOD || u == HTTP_HDR_URL)
+			continue;
+		http_SetH(to, u, ptr);
+		ptr = strchr(ptr, '\0') + 1;
+	}
+	while (*ptr != '\0') {
+		p = strchr(ptr, ':');
+		AN(p);
+		if (!http_findhdr(to, p - ptr, ptr))
+			http_SetHeader(to, ptr);
+		ptr = strchr(ptr, '\0') + 1;
 	}
-	return (l);
 }
 
 /*--------------------------------------------------------------------*/
@@ -702,65 +1069,13 @@ http_FilterReq(struct http *to, const struct http *fm, unsigned how)
 	http_filterfields(to, fm, how);
 }
 
-/*--------------------------------------------------------------------*/
-
-void
-http_FilterResp(const struct http *fm, struct http *to, unsigned how)
-{
-
-	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
-	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
-	to->status = fm->status;
-	http_linkh(to, fm, HTTP_HDR_PROTO);
-	http_linkh(to, fm, HTTP_HDR_STATUS);
-	http_linkh(to, fm, HTTP_HDR_REASON);
-	http_filterfields(to, fm, how);
-}
-
-/*--------------------------------------------------------------------
- * Merge two HTTP headers the "wrong" way. Used by backend IMS to
- * merge in the headers of the validated object with the headers of
- * the 304 response.
- */
-
-void
-http_Merge(const struct http *fm, struct http *to, int not_ce)
-{
-	unsigned u, v;
-	const char *p;
-
-	to->status = fm->status;
-	http_linkh(to, fm, HTTP_HDR_PROTO);
-	http_linkh(to, fm, HTTP_HDR_STATUS);
-	http_linkh(to, fm, HTTP_HDR_REASON);
-
-	for (u = HTTP_HDR_FIRST; u < fm->nhd; u++)
-		fm->hdf[u] |= HDF_MARKER;
-	if (not_ce) {
-		u = http_findhdr(fm,
-		    H_Content_Encoding[0] - 1, H_Content_Encoding + 1);
-		if (u > 0)
-			fm->hdf[u] &= ~HDF_MARKER;
-	}
-	for (v = HTTP_HDR_FIRST; v < to->nhd; v++) {
-		p = strchr(to->hd[v].b, ':');
-		AN(p);
-		u = http_findhdr(fm, p - to->hd[v].b, to->hd[v].b);
-		if (u)
-			fm->hdf[u] &= ~HDF_MARKER;
-	}
-	for (u = HTTP_HDR_FIRST; u < fm->nhd; u++)
-		if (fm->hdf[u] & HDF_MARKER)
-			http_SetHeader(to, fm->hd[u].b);
-}
-
 /*--------------------------------------------------------------------
  * This function copies any header fields which reference foreign
  * storage into our own WS.
  */
 
 void
-http_CopyHome(struct http *hp)
+http_CopyHome(const struct http *hp)
 {
 	unsigned u, l;
 	char *p;
@@ -836,6 +1151,27 @@ http_PrintfHeader(struct http *to, const char *fmt, ...)
 	to->nhd++;
 }
 
+void
+http_TimeHeader(struct http *to, const char *fmt, double now)
+{
+	char *p;
+
+	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
+	p = WS_Alloc(to->ws, strlen(fmt) + VTIM_FORMAT_SIZE);
+	if (p == NULL) {
+		http_fail(to);
+		VSLb(to->vsl, SLT_LostHeader, "%s", fmt);
+		return;
+	}
+	strcpy(p, fmt);
+	VTIM_format(now, strchr(p, '\0'));
+	to->hd[to->nhd].b = p;
+	to->hd[to->nhd].e = strchr(p, '\0');
+	to->hdf[to->nhd] = 0;
+	http_VSLH(to, to->nhd);
+	to->nhd++;
+}
+
 /*--------------------------------------------------------------------*/
 
 void
