@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -31,13 +31,9 @@
 
 #include "config.h"
 
-
-#include <stdio.h>
-#include <stdlib.h>
-
 #include "cache.h"
 
-#include "cache_backend.h"
+#include "cache_director.h"
 #include "hash/hash_slinger.h"
 #include "vav.h"
 #include "vcl.h"
@@ -52,13 +48,11 @@ const void * const vrt_magic_string_unset = &vrt_magic_string_unset;
 /*--------------------------------------------------------------------*/
 
 void
-VRT_error(const struct vrt_ctx *ctx, unsigned code, const char *reason)
+VRT_synth(VRT_CTX, unsigned code, const char *reason)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	VSLb(ctx->vsl, SLT_Debug, "VCL_error(%u, %s)", code,
-	    reason ?  reason : "(null)");
 	if (code < 100 || code > 999)
 		code = 503;
 	ctx->req->err_code = (uint16_t)code;
@@ -69,19 +63,7 @@ VRT_error(const struct vrt_ctx *ctx, unsigned code, const char *reason)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_count(const struct vrt_ctx *ctx, unsigned u)
-{
-
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	if (ctx->vsl != NULL)
-		VSLb(ctx->vsl, SLT_VCL_trace, "%u %u.%u", u,
-		    ctx->vcl->ref[u].line, ctx->vcl->ref[u].pos);
-}
-
-/*--------------------------------------------------------------------*/
-
-void
-VRT_acl_log(const struct vrt_ctx *ctx, const char *msg)
+VRT_acl_log(VRT_CTX, const char *msg)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -90,8 +72,8 @@ VRT_acl_log(const struct vrt_ctx *ctx, const char *msg)
 
 /*--------------------------------------------------------------------*/
 
-static struct http *
-vrt_selecthttp(const struct vrt_ctx *ctx, enum gethdr_e where)
+struct http *
+VRT_selecthttp(VRT_CTX, enum gethdr_e where)
 {
 	struct http *hp;
 
@@ -100,6 +82,9 @@ vrt_selecthttp(const struct vrt_ctx *ctx, enum gethdr_e where)
 	case HDR_REQ:
 		hp = ctx->http_req;
 		break;
+	case HDR_REQ_TOP:
+		hp = ctx->http_req_top;
+		break;
 	case HDR_BEREQ:
 		hp = ctx->http_bereq;
 		break;
@@ -109,25 +94,28 @@ vrt_selecthttp(const struct vrt_ctx *ctx, enum gethdr_e where)
 	case HDR_RESP:
 		hp = ctx->http_resp;
 		break;
-	case HDR_OBJ:
-		hp = ctx->http_obj;
-		break;
 	default:
-		WRONG("vrt_selecthttp 'where' invalid");
+		WRONG("VRT_selecthttp 'where' invalid");
 	}
 	return (hp);
 }
 
 /*--------------------------------------------------------------------*/
 
-char *
-VRT_GetHdr(const struct vrt_ctx *ctx, const struct gethdr_s *hs)
+const char *
+VRT_GetHdr(VRT_CTX, const struct gethdr_s *hs)
 {
-	char *p;
+	const char *p;
 	struct http *hp;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	hp = vrt_selecthttp(ctx, hs->where);
+	if (hs->where == HDR_OBJ) {
+		CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
+		CHECK_OBJ_NOTNULL(ctx->req->objcore, OBJCORE_MAGIC);
+		return(HTTP_GetHdrPack(ctx->req->wrk, ctx->req->objcore,
+		    hs->what));
+	}
+	hp = VRT_selecthttp(ctx, hs->where);
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
 	if (!http_GetHdr(hp, hs->what, &p))
 		return (NULL);
@@ -199,7 +187,7 @@ VRT_String(struct ws *ws, const char *h, const char *p, va_list ap)
  */
 
 const char *
-VRT_CollectString(const struct vrt_ctx *ctx, const char *p, ...)
+VRT_CollectString(VRT_CTX, const char *p, ...)
 {
 	va_list ap;
 	const char *b;
@@ -215,7 +203,7 @@ VRT_CollectString(const struct vrt_ctx *ctx, const char *p, ...)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_SetHdr(const struct vrt_ctx *ctx , const struct gethdr_s *hs,
+VRT_SetHdr(VRT_CTX , const struct gethdr_s *hs,
     const char *p, ...)
 {
 	struct http *hp;
@@ -225,7 +213,7 @@ VRT_SetHdr(const struct vrt_ctx *ctx , const struct gethdr_s *hs,
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	AN(hs);
 	AN(hs->what);
-	hp = vrt_selecthttp(ctx, hs->where);
+	hp = VRT_selecthttp(ctx, hs->where);
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
 	va_start(ap, p);
 	if (p == vrt_magic_string_unset) {
@@ -245,7 +233,7 @@ VRT_SetHdr(const struct vrt_ctx *ctx , const struct gethdr_s *hs,
 /*--------------------------------------------------------------------*/
 
 void
-VRT_handling(const struct vrt_ctx *ctx, unsigned hand)
+VRT_handling(VRT_CTX, unsigned hand)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -258,34 +246,34 @@ VRT_handling(const struct vrt_ctx *ctx, unsigned hand)
  */
 
 void
-VRT_hashdata(const struct vrt_ctx *ctx, const char *str, ...)
+VRT_hashdata(VRT_CTX, const char *str, ...)
 {
 	va_list ap;
 	const char *p;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-	HSH_AddString(ctx->req, str);
+	AN(ctx->specific);
+	HSH_AddString(ctx->req, ctx->specific, str);
 	va_start(ap, str);
 	while (1) {
 		p = va_arg(ap, const char *);
 		if (p == vrt_magic_string_end)
 			break;
-		HSH_AddString(ctx->req, p);
-		VSLb(ctx->vsl, SLT_Hash, "%s", str);
+		HSH_AddString(ctx->req, ctx->specific, p);
 	}
 	va_end(ap);
 	/*
 	 * Add a 'field-separator' to make it more difficult to
 	 * manipulate the hash.
 	 */
-	HSH_AddString(ctx->req, NULL);
+	HSH_AddString(ctx->req, ctx->specific, NULL);
 }
 
 /*--------------------------------------------------------------------*/
 
 double
-VRT_r_now(const struct vrt_ctx *ctx)
+VRT_r_now(VRT_CTX)
 {
 
 	(void)ctx;
@@ -295,7 +283,7 @@ VRT_r_now(const struct vrt_ctx *ctx)
 /*--------------------------------------------------------------------*/
 
 char *
-VRT_IP_string(const struct vrt_ctx *ctx, VCL_IP ip)
+VRT_IP_string(VRT_CTX, VCL_IP ip)
 {
 	char *p;
 	unsigned len;
@@ -315,7 +303,7 @@ VRT_IP_string(const struct vrt_ctx *ctx, VCL_IP ip)
 }
 
 char *
-VRT_INT_string(const struct vrt_ctx *ctx, long num)
+VRT_INT_string(VRT_CTX, long num)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -323,7 +311,7 @@ VRT_INT_string(const struct vrt_ctx *ctx, long num)
 }
 
 char *
-VRT_REAL_string(const struct vrt_ctx *ctx, double num)
+VRT_REAL_string(VRT_CTX, double num)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -331,7 +319,7 @@ VRT_REAL_string(const struct vrt_ctx *ctx, double num)
 }
 
 char *
-VRT_TIME_string(const struct vrt_ctx *ctx, double t)
+VRT_TIME_string(VRT_CTX, double t)
 {
 	char *p;
 
@@ -361,7 +349,7 @@ VRT_BOOL_string(unsigned val)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_Rollback(const struct vrt_ctx *ctx, const struct http *hp)
+VRT_Rollback(VRT_CTX, const struct http *hp)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -382,22 +370,13 @@ VRT_Rollback(const struct vrt_ctx *ctx, const struct http *hp)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_synth_page(const struct vrt_ctx *ctx, const char *str, ...)
+VRT_synth_page(VRT_CTX, const char *str, ...)
 {
 	va_list ap;
 	const char *p;
 	struct vsb *vsb;
 
-	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
-	if (ctx->method == VCL_MET_BACKEND_ERROR) {
-		CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
-		vsb = ctx->bo->synth_body;
-	} else {
-		CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
-		vsb = ctx->req->synth_body;
-	}
-	AN(vsb);
-
+	CAST_OBJ_NOTNULL(vsb, ctx->specific, VSB_MAGIC);
 	va_start(ap, str);
 	p = str;
 	while (p != vrt_magic_string_end) {
@@ -412,19 +391,20 @@ VRT_synth_page(const struct vrt_ctx *ctx, const char *str, ...)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_ban_string(const struct vrt_ctx *ctx, const char *str)
+VRT_ban_string(VRT_CTX, const char *str)
 {
 	char *a1, *a2, *a3;
 	char **av;
-	struct ban *b;
+	struct ban_proto *bp;
+	const char *err;
 	int i;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	AN(ctx->vsl);
 	AN(str);
 
-	b = BAN_New();
-	if (b == NULL) {
+	bp = BAN_Build();
+	if (bp == NULL) {
 		VSLb(ctx->vsl, SLT_VCL_Error, "ban(): Out of Memory");
 		return;
 	}
@@ -433,7 +413,7 @@ VRT_ban_string(const struct vrt_ctx *ctx, const char *str)
 	if (av[0] != NULL) {
 		VSLb(ctx->vsl, SLT_VCL_Error, "ban(): %s", av[0]);
 		VAV_Free(av);
-		BAN_Free(b);
+		BAN_Abandon(bp);
 		return;
 	}
 	for (i = 0; ;) {
@@ -455,13 +435,17 @@ VRT_ban_string(const struct vrt_ctx *ctx, const char *str)
 			    "ban(): Expected second operand.");
 			break;
 		}
-		if (BAN_AddTest(b, a1, a2, a3) || av[++i] == NULL) {
-			a1 = BAN_Insert(b);
-			if (a1 != NULL) {
-				VSLb(ctx->vsl, SLT_VCL_Error,
-				    "ban(): %s", a1);
-				BAN_Free_Errormsg(a1);
-			}
+		err = BAN_AddTest(bp, a1, a2, a3);
+		if (err) {
+			VSLb(ctx->vsl, SLT_VCL_Error, "ban(): %s", err);
+			break;
+		}
+		if (av[++i] == NULL) {
+			err = BAN_Commit(bp);
+			if (err == NULL)
+				bp = NULL;
+			else
+				VSLb(ctx->vsl, SLT_VCL_Error, "ban(): %s", err);
 			break;
 		}
 		if (strcmp(av[i], "&&")) {
@@ -471,15 +455,19 @@ VRT_ban_string(const struct vrt_ctx *ctx, const char *str)
 			break;
 		}
 	}
+	if (bp != NULL)
+		BAN_Abandon(bp);
 	VAV_Free(av);
 }
 
 /*--------------------------------------------------------------------
  *
+ * XXX this really should be ssize_t VRT_CacheReqBody(VRT_CTX, size_t)
+ * - change with next VRT major bump
  */
 
 int
-VRT_CacheReqBody(const struct vrt_ctx *ctx, long long maxsize)
+VRT_CacheReqBody(VRT_CTX, long long maxsize)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
@@ -487,9 +475,9 @@ VRT_CacheReqBody(const struct vrt_ctx *ctx, long long maxsize)
 	if (ctx->method != VCL_MET_RECV) {
 		VSLb(ctx->vsl, SLT_VCL_Error,
 		    "req.body can only be cached in vcl_recv{}");
-		return (0);
+		return (-1);
 	}
-	return (HTTP1_CacheReqBody(ctx->req, maxsize));
+	return (VRB_Cache(ctx->req, maxsize));
 }
 
 /*--------------------------------------------------------------------
@@ -497,30 +485,20 @@ VRT_CacheReqBody(const struct vrt_ctx *ctx, long long maxsize)
  */
 
 void
-VRT_purge(const struct vrt_ctx *ctx, double ttl, double grace, double keep)
+VRT_purge(VRT_CTX, double ttl, double grace, double keep)
 {
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req, REQ_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->req->wrk, WORKER_MAGIC);
 	if (ctx->method == VCL_MET_HIT)
-		HSH_Purge(ctx->req->wrk, ctx->req->obj->objcore->objhead,
+		HSH_Purge(ctx->req->wrk, ctx->req->objcore->objhead,
 		    ttl, grace, keep);
 	else if (ctx->method == VCL_MET_MISS)
 		HSH_Purge(ctx->req->wrk, ctx->req->objcore->objhead,
 		    ttl, grace, keep);
 }
 
-/*--------------------------------------------------------------------
- */
-
-void
-VRT_priv_fini(const struct vmod_priv *p)
-{
-	if (p->priv != (void*)0 && p->free != (void*)0)
-		p->free(p->priv);
-}
-
 /*--------------------------------------------------------------------
  * Simple stuff
  */
