@@ -31,7 +31,6 @@
 #include "config.h"
 
 #include <stdio.h>
-#include <stdarg.h>
 
 #include "cache.h"
 
@@ -57,6 +56,7 @@ WS_Assert(const struct ws *ws)
 		assert(ws->r <= ws->e);
 		assert(PAOK(ws->r));
 	}
+	assert(*ws->e == 0x15);
 }
 
 /*
@@ -71,14 +71,14 @@ WS_Init(struct ws *ws, const char *id, void *space, unsigned len)
 	DSL(DBG_WORKSPACE, 0,
 	    "WS_Init(%p, \"%s\", %p, %u)", ws, id, space, len);
 	assert(space != NULL);
-	memset(ws, 0, sizeof *ws);
-	ws->magic = WS_MAGIC;
+	INIT_OBJ(ws, WS_MAGIC);
 	ws->s = space;
 	assert(PAOK(space));
-	len = PRNDDN(len);
+	len = PRNDDN(len - 1);
 	ws->e = ws->s + len;
+	*ws->e = 0x15;
 	ws->f = ws->s;
-	assert(id[0] & 0x40);
+	assert(id[0] & 0x20);
 	assert(strlen(id) < sizeof ws->id);
 	strcpy(ws->id, id);
 	WS_Assert(ws);
@@ -90,7 +90,15 @@ WS_MarkOverflow(struct ws *ws)
 {
 	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
 
-	ws->id[0] &= ~0x40;		// Cheasy toupper()
+	ws->id[0] &= ~0x20;		// cheesy toupper()
+}
+
+static void
+ws_ClearOverflow(struct ws *ws)
+{
+	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
+
+	ws->id[0] |= 0x20;		// cheesy tolower()
 }
 
 /*
@@ -111,10 +119,11 @@ WS_Reset(struct ws *ws, char *p)
 		assert(p < ws->e);
 		ws->f = p;
 	}
+	ws_ClearOverflow(ws);
 	WS_Assert(ws);
 }
 
-char *
+void *
 WS_Alloc(struct ws *ws, unsigned bytes)
 {
 	char *r;
@@ -178,6 +187,7 @@ WS_Printf(struct ws *ws, const char *fmt, ...)
 	va_end(ap);
 	if (v >= u) {
 		WS_Release(ws, 0);
+		WS_MarkOverflow(ws);
 		p = NULL;
 	} else {
 		WS_Release(ws, v + 1);
@@ -246,7 +256,7 @@ WS_Overflowed(const struct ws *ws)
 {
 	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
 
-	if (ws->id[0] & 0x40)
+	if (ws->id[0] & 0x20)
 		return (0);
 	return (1);
 }
