@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -66,16 +66,14 @@ void HSH_Cleanup(struct worker *w);
 enum lookup_e HSH_Lookup(struct req *, struct objcore **, struct objcore **,
     int wait_for_busy, int always_insert);
 void HSH_Ref(struct objcore *o);
-void HSH_Drop(struct worker *, struct object **);
 void HSH_Init(const struct hash_slinger *slinger);
-void HSH_AddString(const struct req *, const char *str);
+void HSH_AddString(struct req *, void *ctx, const char *str);
 void HSH_Insert(struct worker *, const void *hash, struct objcore *);
 void HSH_Purge(struct worker *, struct objhead *, double ttl, double grace,
     double keep);
 void HSH_config(const char *h_arg);
 struct busyobj *HSH_RefBusy(const struct objcore *oc);
 struct objcore *HSH_Private(struct worker *wrk);
-struct objcore *HSH_NewObjCore(struct worker *wrk);
 
 #ifdef VARNISH_CACHE_CHILD
 
@@ -110,12 +108,11 @@ struct objhead {
 };
 
 void HSH_Fail(struct objcore *);
-void HSH_Unbusy(struct dstat *, struct objcore *);
+void HSH_Unbusy(struct worker *, struct objcore *);
 void HSH_Complete(struct objcore *oc);
-void HSH_DeleteObjHead(struct dstat *, struct objhead *oh);
-int HSH_DerefObjHead(struct dstat *, struct objhead **poh);
-int HSH_DerefObjCore(struct dstat *, struct objcore **ocp);
-int HSH_DerefObj(struct dstat *, struct object **o);
+void HSH_DeleteObjHead(struct worker *, struct objhead *oh);
+int HSH_DerefObjHead(struct worker *, struct objhead **poh);
+int HSH_DerefObjCore(struct worker *, struct objcore **ocp);
 #endif /* VARNISH_CACHE_CHILD */
 
 extern const struct hash_slinger hsl_slinger;
