@@ -45,7 +45,6 @@
 #include <stdio.h>
 
 #include "mgt/mgt.h"
-#include "common/params.h"
 
 #include "mgt/mgt_param.h"
 
@@ -110,9 +109,7 @@ struct parspec WRK_parspec[] = {
 		"\n"
 		"Do not set this higher than you have to, since excess "
 		"worker threads soak up RAM and CPU and generally just get "
-		"in the way of getting work done.\n"
-		"\n"
-		"Minimum is 10 threads.",
+		"in the way of getting work done.",
 		DELAYED_EFFECT,
 		"5000", "threads" },
 	{ "thread_pool_min", tweak_thread_pool_min, &mgt_param.wthread_min,
@@ -131,9 +128,7 @@ struct parspec WRK_parspec[] = {
 		"Thread idle threshold.\n"
 		"\n"
 		"Threads in excess of thread_pool_min, which have been idle "
-		"for at least this long, will be destroyed.\n"
-		"\n"
-		"Minimum is 10 seconds.",
+		"for at least this long, will be destroyed.",
 		EXPERIMENTAL | DELAYED_EFFECT,
 		"300", "seconds" },
 	{ "thread_pool_destroy_delay",
@@ -141,9 +136,7 @@ struct parspec WRK_parspec[] = {
 		"0.01", NULL,
 		"Wait this long after destroying a thread.\n"
 		"\n"
-		"This controls the decay of thread pools when idle(-ish).\n"
-		"\n"
-		"Minimum is 0.01 second.",
+		"This controls the decay of thread pools when idle(-ish).",
 		EXPERIMENTAL | DELAYED_EFFECT,
 		"1", "seconds" },
 	{ "thread_pool_add_delay",
@@ -198,15 +191,6 @@ struct parspec WRK_parspec[] = {
 		"be dropped instead of queued.",
 		EXPERIMENTAL,
 		"20", "" },
-	{ "rush_exponent", tweak_uint, &mgt_param.rush_exponent,
-		"2", NULL,
-		"How many parked request we start for each completed "
-		"request on the object.\n"
-		"NB: Even with the implict delay of delivery, "
-		"this parameter controls an exponential increase in "
-		"number of worker threads.",
-		EXPERIMENTAL,
-		"3", "requests per request" },
 	{ "thread_pool_stack",
 		tweak_bytes, &mgt_param.wthread_stacksize,
 		NULL, NULL,
