@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -46,6 +46,7 @@
 #include <string.h>
 #include <unistd.h>
 
+#include "vdef.h"
 #include "vcurses.h"
 #include "vapi/vsl.h"
 #include "vapi/vsm.h"
@@ -193,7 +194,8 @@ static int /*__match_proto__ (VSLQ_dispatch_f)*/
 accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	void *priv)
 {
-	int i, j, tag, skip, match, hit;
+	int i, tag, skip, match, hit;
+	unsigned u;
 	double value;
 	struct VSL_transaction *tr;
 
@@ -259,13 +261,15 @@ accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 
 		/* phase out old data */
 		if (nhist == HIST_N) {
-			j = rr_hist[next_hist];
-			if (j < 0) {
-				assert(bucket_miss[-j] > 0);
-				bucket_miss[-j]--;
+			u = rr_hist[next_hist];
+			if (u >= hist_buckets) {
+				u -= hist_buckets;
+				assert(u < hist_buckets);
+				assert(bucket_hit[u] > 0);
+				bucket_hit[u]--;
 			} else {
-				assert(bucket_hit[j] > 0);
-				bucket_hit[j]--;
+				assert(bucket_miss[u] > 0);
+				bucket_miss[u]--;
 			}
 		} else {
 			++nhist;
@@ -274,10 +278,10 @@ accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 		/* phase in new data */
 		if (hit) {
 			bucket_hit[i]++;
-			rr_hist[next_hist] = i;
+			rr_hist[next_hist] = i + hist_buckets;
 		} else {
 			bucket_miss[i]++;
-			rr_hist[next_hist] = -i;
+			rr_hist[next_hist] = i;
 		}
 		if (++next_hist == HIST_N) {
 			next_hist = 0;
@@ -287,6 +291,12 @@ accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	return (0);
 }
 
+static int __match_proto__(VUT_cb_f)
+sighup(void)
+{
+	return (1);
+}
+
 static void *
 do_curses(void *arg)
 {
@@ -461,6 +471,7 @@ main(int argc, char **argv)
 	}
 	VUT.dispatch_f = &accumulate;
 	VUT.dispatch_priv = NULL;
+	VUT.sighup_f = sighup;
 	VUT_Main();
 	end_of_file = 1;
 	AZ(pthread_join(thr, NULL));
