@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -44,6 +44,7 @@
 #include "vapi/vsl.h"
 #include "vapi/voptget.h"
 #include "vas.h"
+#include "vdef.h"
 #include "vpf.h"
 #include "vsb.h"
 #include "vut.h"
@@ -54,7 +55,7 @@ static const char progname[] = "varnishlog";
 struct log {
 	/* Options */
 	int		a_opt;
-	int		B_opt;
+	int		A_opt;
 	char		*w_arg;
 
 	/* State */
@@ -77,13 +78,13 @@ openout(int append)
 {
 
 	AN(LOG.w_arg);
-	if (LOG.B_opt)
-		LOG.fo = VSL_WriteOpen(VUT.vsl, LOG.w_arg, append, 0);
-	else
+	if (LOG.A_opt)
 		LOG.fo = fopen(LOG.w_arg, append ? "a" : "w");
+	else
+		LOG.fo = VSL_WriteOpen(VUT.vsl, LOG.w_arg, append, 0);
 	if (LOG.fo == NULL)
 		VUT_Error(2, "Can't open output file (%s)",
-		    LOG.B_opt ? VSL_Error(VUT.vsl) : strerror(errno));
+		    LOG.A_opt ? strerror(errno) : VSL_Error(VUT.vsl));
 	VUT.dispatch_priv = LOG.fo;
 }
 
@@ -109,6 +110,12 @@ flushout(void)
 	return (0);
 }
 
+static int __match_proto__(VUT_cb_f)
+sighup(void)
+{
+	return (1);
+}
+
 int
 main(int argc, char * const *argv)
 {
@@ -123,9 +130,9 @@ main(int argc, char * const *argv)
 			/* Append to file */
 			LOG.a_opt = 1;
 			break;
-		case 'B':
-			/* Binary output */
-			LOG.B_opt = 1;
+		case 'A':
+			/* Text output */
+			LOG.A_opt = 1;
 			break;
 		case 'h':
 			/* Usage help */
@@ -144,15 +151,20 @@ main(int argc, char * const *argv)
 	if (optind != argc)
 		usage(1);
 
+	if (VUT.D_opt && !LOG.w_arg)
+		VUT_Error(1, "Missing -w option");
+
 	/* Setup output */
-	if (LOG.B_opt)
-		VUT.dispatch_f = VSL_WriteTransactions;
-	else
+	if (LOG.A_opt || !LOG.w_arg)
 		VUT.dispatch_f = VSL_PrintTransactions;
+	else
+		VUT.dispatch_f = VSL_WriteTransactions;
+	VUT.sighup_f = sighup;
 	if (LOG.w_arg) {
 		openout(LOG.a_opt);
 		AN(LOG.fo);
-		VUT.sighup_f = rotateout;
+		if (VUT.D_opt)
+			VUT.sighup_f = rotateout;
 	} else
 		LOG.fo = stdout;
 	VUT.idle_f = flushout;
