@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Anders Berg <andersb@vgnett.no>
@@ -41,6 +41,8 @@
 
 #include "config.h"
 
+#define _WITH_GETLINE
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <unistd.h>
@@ -57,6 +59,7 @@
 #include "vapi/vsl.h"
 #include "vapi/voptget.h"
 #include "vas.h"
+#include "vdef.h"
 #include "vcs.h"
 #include "vsb.h"
 #include "vut.h"
@@ -99,12 +102,12 @@ struct format {
 	unsigned		magic;
 #define FORMAT_MAGIC		0xC3119CDA
 
+	char			time_type;
 	VTAILQ_ENTRY(format)	list;
 	format_f		*func;
 	struct fragment		*frag;
 	char			*string;
 	const char *const	*strptr;
-	char			time_type;
 	char			*time_fmt;
 };
 
@@ -182,16 +185,61 @@ flushout(void)
 	return (0);
 }
 
+static int
+vsb_esc_cat(struct vsb *sb, const char *b, const char *e)
+{
+	AN(b);
+
+	for (; b < e; b++) {
+		if (isspace(*b)) {
+			switch (*b) {
+			case '\n':
+				VSB_cat(sb, "\\n");
+				break;
+			case '\t':
+				VSB_cat(sb, "\\t");
+				break;
+			case '\f':
+				VSB_cat(sb, "\\f");
+				break;
+			case '\r':
+				VSB_cat(sb, "\\r");
+				break;
+			case '\v':
+				VSB_cat(sb, "\\v");
+				break;
+			default:
+				VSB_putc(sb, *b);
+				break;
+			}
+		} else if (isprint(*b)) {
+			switch (*b) {
+			case '"':
+				VSB_cat(sb, "\\\"");
+				break;
+			case '\\':
+				VSB_cat(sb, "\\\\");
+				break;
+			default:
+				VSB_putc(sb, *b);
+				break;
+			}
+		} else
+			VSB_printf(sb, "\\x%02hhx", *b);
+	}
+
+	return (VSB_error(sb));
+}
+
 static inline int
 vsb_fcat(struct vsb *vsb, const struct fragment *f, const char *dflt)
 {
-
 	if (f->gen == CTX.gen) {
 		assert(f->b <= f->e);
-		return (VSB_bcat(vsb, f->b, f->e - f->b));
+		return (vsb_esc_cat(vsb, f->b, f->e));
 	}
 	if (dflt)
-		return (VSB_cat(vsb, dflt));
+		return (vsb_esc_cat(vsb, dflt, dflt + strlen(dflt)));
 	return (-1);
 }
 
@@ -226,10 +274,10 @@ format_fragment(const struct format *format)
 	if (format->frag->gen != CTX.gen) {
 		if (format->string == NULL)
 			return (-1);
-		AZ(VSB_cat(CTX.vsb, format->string));
+		AZ(vsb_esc_cat(CTX.vsb, format->string,
+		    format->string + strlen(format->string)));
 		return (0);
 	}
-
 	AZ(vsb_fcat(CTX.vsb, format->frag, NULL));
 	return (1);
 }
@@ -318,13 +366,14 @@ format_auth(const struct format *format)
 		CTX.frag[F_auth].e)) {
 		if (format->string == NULL)
 			return (-1);
-		AZ(VSB_cat(CTX.vsb, format->string));
+		AZ(vsb_esc_cat(CTX.vsb, format->string,
+			format->string + strlen(format->string)));
 		return (0);
 	}
 	q = strchr(buf, ':');
 	if (q != NULL)
 		*q = '\0';
-	AZ(VSB_cat(CTX.vsb, buf));
+	AZ(vsb_esc_cat(CTX.vsb, buf, buf + strlen(buf)));
 	return (1);
 }
 
@@ -482,6 +531,7 @@ addf_auth(const char *str)
 	struct format *f;
 
 	ALLOC_OBJ(f, FORMAT_MAGIC);
+	AN(f);
 	f->func = &format_auth;
 	if (str != NULL) {
 		f->string = strdup(str);
@@ -653,7 +703,7 @@ isprefix(const char *str, const char *prefix, const char *end,
 }
 
 static void
-frag_fields(const char *b, const char *e, ...)
+frag_fields(int force, const char *b, const char *e, ...)
 {
 	va_list ap;
 	const char *p, *q;
@@ -680,7 +730,7 @@ frag_fields(const char *b, const char *e, ...)
 			q++;
 
 		if (field == n) {
-			if (frag->gen != CTX.gen) {
+			if (frag->gen != CTX.gen || force) {
 				/* We only grab the same matching field once */
 				frag->gen = CTX.gen;
 				frag->b = p;
@@ -696,10 +746,10 @@ frag_fields(const char *b, const char *e, ...)
 }
 
 static void
-frag_line(const char *b, const char *e, struct fragment *f)
+frag_line(int force, const char *b, const char *e, struct fragment *f)
 {
 
-	if (f->gen == CTX.gen)
+	if (f->gen == CTX.gen && !force)
 		/* We only grab the same matching record once */
 		return;
 
@@ -724,7 +774,7 @@ process_hdr(const struct watch_head *head, const char *b, const char *e)
 	VTAILQ_FOREACH(w, head, list) {
 		if (strncasecmp(b, w->key, w->keylen))
 			continue;
-		frag_line(b + w->keylen, e, &w->frag);
+		frag_line(1, b + w->keylen, e, &w->frag);
 	}
 }
 
@@ -737,7 +787,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	const char *b, *e, *p;
 	struct watch *w;
 	int i, skip;
-
 	(void)vsl;
 	(void)priv;
 
@@ -764,32 +813,34 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				skip = 1;
 				break;
 			case SLT_PipeAcct:
-				frag_fields(b, e,
+				frag_fields(0, b, e,
 				    3, &CTX.frag[F_I],
 				    4, &CTX.frag[F_O],
 				    0, NULL);
 				break;
 			case SLT_ReqStart:
-				frag_fields(b, e, 1, &CTX.frag[F_h], 0, NULL);
+				frag_fields(0, b, e,
+				    1, &CTX.frag[F_h],
+				    0, NULL);
 				break;
 			case SLT_ReqMethod:
-				frag_line(b, e, &CTX.frag[F_m]);
+				frag_line(0, b, e, &CTX.frag[F_m]);
 				break;
 			case SLT_ReqURL:
 				p = memchr(b, '?', e - b);
 				if (p == NULL)
 					p = e;
-				frag_line(b, p, &CTX.frag[F_U]);
-				frag_line(p, e, &CTX.frag[F_q]);
+				frag_line(0, b, p, &CTX.frag[F_U]);
+				frag_line(0, p, e, &CTX.frag[F_q]);
 				break;
 			case SLT_ReqProtocol:
-				frag_line(b, e, &CTX.frag[F_H]);
+				frag_line(0, b, e, &CTX.frag[F_H]);
 				break;
 			case SLT_RespStatus:
-				frag_line(b, e, &CTX.frag[F_s]);
+				frag_line(1, b, e, &CTX.frag[F_s]);
 				break;
 			case SLT_ReqAcct:
-				frag_fields(b, e,
+				frag_fields(0, b, e,
 				    3, &CTX.frag[F_I],
 				    5, &CTX.frag[F_b],
 				    6, &CTX.frag[F_O],
@@ -797,26 +848,26 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				break;
 			case SLT_Timestamp:
 				if (isprefix(b, "Start:", e, &p)) {
-					frag_fields(p, e, 1,
+					frag_fields(0, p, e, 1,
 					    &CTX.frag[F_tstart], 0, NULL);
 
 				} else if (isprefix(b, "Resp:", e, &p) ||
 					isprefix(b, "PipeSess:", e, &p)) {
-					frag_fields(p, e, 1,
+					frag_fields(0, p, e, 1,
 					    &CTX.frag[F_tend], 0, NULL);
 
 				} else if (isprefix(b, "Process:", e, &p) ||
 					isprefix(b, "Pipe:", e, &p)) {
-					frag_fields(p, e, 2,
+					frag_fields(0, p, e, 2,
 					    &CTX.frag[F_ttfb], 0, NULL);
 				}
 				break;
 			case SLT_ReqHeader:
 				if (isprefix(b, "Host:", e, &p))
-					frag_line(p, e, &CTX.frag[F_host]);
+					frag_line(0, p, e, &CTX.frag[F_host]);
 				else if (isprefix(b, "Authorization:", e, &p) &&
 				    isprefix(p, "basic", e, &p))
-					frag_line(p, e, &CTX.frag[F_auth]);
+					frag_line(0, p, e, &CTX.frag[F_auth]);
 				break;
 			case SLT_VCL_call:
 				if (!strcasecmp(b, "recv")) {
@@ -831,12 +882,12 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 				} else if (!strcasecmp(b, "pass")) {
 					CTX.hitmiss = "miss";
 					CTX.handling = "pass";
-				} else if (!strcasecmp(b, "error")) {
-					/* Arguably, error isn't a hit or
+				} else if (!strcasecmp(b, "synth")) {
+					/* Arguably, synth isn't a hit or
 					   a miss, but miss is less
 					   wrong */
 					CTX.hitmiss = "miss";
-					CTX.handling = "error";
+					CTX.handling = "synth";
 				}
 				break;
 			case SLT_VCL_return:
@@ -850,7 +901,6 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 			default:
 				break;
 			}
-
 			if (tag == SLT_VCL_Log) {
 				VTAILQ_FOREACH(w, &CTX.watch_vcl_log, list) {
 					CHECK_OBJ_NOTNULL(w, WATCH_MAGIC);
@@ -862,7 +912,7 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 					p++;
 					if (p > e)
 						continue;
-					frag_line(p, e, &w->frag);
+					frag_line(0, p, e, &w->frag);
 				}
 			}
 			if (tag == SLT_ReqHeader)
@@ -880,6 +930,34 @@ dispatch_f(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	return (0);
 }
 
+static int __match_proto__(VUT_cb_f)
+sighup(void)
+{
+	return (1);
+}
+
+static char *
+read_format(const char *formatfile)
+{
+	FILE *fmtfile;
+	size_t len = 0;
+	char *fmt = NULL;
+
+	fmtfile = fopen(formatfile, "r");
+	if (fmtfile == NULL)
+		VUT_Error(1, "Can't open format file (%s)", strerror(errno));
+	if (getline(&fmt, &len, fmtfile) == -1) {
+		free(fmt);
+		if (feof(fmtfile))
+			VUT_Error(1, "Empty format file");
+		else
+			VUT_Error(1, "Can't read format from file (%s)",
+			    strerror(errno));
+	}
+	fclose(fmtfile);
+	return (fmt);
+}
+
 int
 main(int argc, char * const *argv)
 {
@@ -909,6 +987,13 @@ main(int argc, char * const *argv)
 			format = strdup(optarg);
 			AN(format);
 			break;
+		case 'f':
+			/* Format string from file */
+			if (format != NULL)
+				free(format);
+			format = read_format(optarg);
+			AN(format);
+			break;
 		case 'h':
 			/* Usage help */
 			usage(0);
@@ -925,6 +1010,9 @@ main(int argc, char * const *argv)
 	if (optind != argc)
 		usage(1);
 
+	if (VUT.D_opt && !CTX.w_arg)
+		VUT_Error(1, "Missing -w option");
+
 	/* Check for valid grouping mode */
 	assert(VUT.g_arg < VSL_g__MAX);
 	if (VUT.g_arg != VSL_g_vxid && VUT.g_arg != VSL_g_request)
@@ -941,10 +1029,12 @@ main(int argc, char * const *argv)
 	/* Setup output */
 	VUT.dispatch_f = &dispatch_f;
 	VUT.dispatch_priv = NULL;
+	VUT.sighup_f = sighup;
 	if (CTX.w_arg) {
 		openout(CTX.a_opt);
 		AN(CTX.fo);
-		VUT.sighup_f = &rotateout;
+		if (VUT.D_opt)
+			VUT.sighup_f = &rotateout;
 	} else
 		CTX.fo = stdout;
 	VUT.idle_f = &flushout;
