@@ -27,7 +27,6 @@
  *
  */
 
-#include <errno.h>
 #include <limits.h>
 #include <signal.h>
 #include <stdint.h>
@@ -55,7 +54,7 @@ struct cmds {
 	cmd_f		*cmd;
 };
 
-void parse_string(char *buf, const struct cmds *cmd, void *priv,
+void parse_string(const char *spec, const struct cmds *cmd, void *priv,
     struct vtclog *vl);
 
 cmd_f cmd_delay;
@@ -64,14 +63,18 @@ cmd_f cmd_client;
 cmd_f cmd_varnish;
 cmd_f cmd_sema;
 cmd_f cmd_logexp;
+cmd_f cmd_process;
 
 extern volatile sig_atomic_t vtc_error; /* Error, bail out */
 extern int vtc_stop;		/* Abandon current test, no error */
 extern pthread_t	vtc_thread;
 extern int iflg;
 extern unsigned vtc_maxdur;
+extern int vtc_witness;
+extern int feature_dns;
 
 void init_sema(void);
+void init_server(void);
 
 int http_process(struct vtclog *vl, const char *spec, int sock, int *sfd);
 
@@ -81,7 +84,7 @@ void vtc_loginit(char *buf, unsigned buflen);
 struct vtclog *vtc_logopen(const char *id);
 void vtc_logclose(struct vtclog *vl);
 void vtc_log(struct vtclog *vl, int lvl, const char *fmt, ...)
-    __printflike(3, 4);
+    __v_printflike(3, 4);
 void vtc_dump(struct vtclog *vl, int lvl, const char *pfx,
     const char *str, int len);
 void vtc_hexdump(struct vtclog *vl, int lvl, const char *pfx,
@@ -93,8 +96,8 @@ int exec_file(const char *fn, const char *script, const char *tmpdir,
 void macro_undef(struct vtclog *vl, const char *instance, const char *name);
 void macro_def(struct vtclog *vl, const char *instance, const char *name,
     const char *fmt, ...)
-    __printflike(4, 5);
+    __v_printflike(4, 5);
 struct vsb *macro_expand(struct vtclog *vl, const char *text);
 
 void extmacro_def(const char *name, const char *fmt, ...)
-    __printflike(2, 3);
+    __v_printflike(2, 3);
