@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -49,6 +49,7 @@
 #include "vapi/vsl.h"
 #include "vapi/voptget.h"
 #include "vas.h"
+#include "vdef.h"
 #include "vcs.h"
 #include "vtree.h"
 #include "vsb.h"
@@ -108,10 +109,10 @@ static int f_flag = 0;
 
 static unsigned maxfieldlen = 0;
 
-VRB_PROTOTYPE(t_order, top, e_order, cmp_order);
-VRB_GENERATE(t_order, top, e_order, cmp_order);
-VRB_PROTOTYPE(t_key, top, e_key, cmp_key);
-VRB_GENERATE(t_key, top, e_key, cmp_key);
+VRB_PROTOTYPE_STATIC(t_order, top, e_order, cmp_order);
+VRB_GENERATE_STATIC(t_order, top, e_order, cmp_order);
+VRB_PROTOTYPE_STATIC(t_key, top, e_key, cmp_key);
+VRB_GENERATE_STATIC(t_key, top, e_key, cmp_key);
 
 static int __match_proto__(VSLQ_dispatch_f)
 accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
@@ -179,6 +180,12 @@ accumulate(struct VSL_data *vsl, struct VSL_transaction * const pt[],
 	return (0);
 }
 
+static int __match_proto__(VUT_cb_f)
+sighup(void)
+{
+	return (1);
+}
+
 static void
 update(int p)
 {
@@ -361,6 +368,7 @@ main(int argc, char **argv)
 	}
 	VUT.dispatch_f = &accumulate;
 	VUT.dispatch_priv = NULL;
+	VUT.sighup_f = sighup;
 	VUT_Main();
 	end_of_file = 1;
 	if (once)
