@@ -28,8 +28,6 @@
  *
  */
 
-#include <stdint.h>
-
 /* from libvarnish/vct.c */
 
 #define VCT_SP			(1<<0)
@@ -42,19 +40,20 @@
 #define VCT_HEX			(1<<6)
 #define VCT_XMLNAMESTART	(1<<7)
 #define VCT_XMLNAME		(1<<8)
+#define VCT_TCHAR		(1<<9)
 
 extern const uint16_t vct_typtab[256];
 
 static inline int
-vct_is(unsigned char x, uint16_t y)
+vct_is(int x, uint16_t y)
 {
 
+	x &= 0xff;
 	return (vct_typtab[x] & (y));
 }
 
 #define vct_issp(x) vct_is(x, VCT_SP)
 #define vct_ishex(x) vct_is(x, VCT_HEX)
-#define vct_iscrlf(x) vct_is(x, VCT_CRLF)
 #define vct_islws(x) vct_is(x, VCT_LWS)
 #define vct_isctl(x) vct_is(x, VCT_CTL)
 #define vct_isdigit(x) vct_is(x, VCT_DIGIT)
@@ -63,6 +62,9 @@ vct_is(unsigned char x, uint16_t y)
 #define vct_issepctl(x) vct_is(x, VCT_SEPARATOR | VCT_CTL)
 #define vct_isxmlnamestart(x) vct_is(x, VCT_XMLNAMESTART)
 #define vct_isxmlname(x) vct_is(x, VCT_XMLNAMESTART | VCT_XMLNAME)
+#define vct_istchar(x) vct_is(x, VCT_ALPHA | VCT_DIGIT | VCT_TCHAR)
+
+#define vct_iscrlf(p) (((p)[0] == '\r' && (p)[1] == '\n') || (p)[0] == '\n')
 
 /* NB: VCT always operate in ASCII, don't replace 0x0d with \r etc. */
-#define vct_skipcrlf(p) (p[0] == 0x0d && p[1] == 0x0a ? 2 : 1)
+#define vct_skipcrlf(p) ((p)[0] == 0x0d && (p)[1] == 0x0a ? 2 : 1)
