@@ -27,11 +27,8 @@
  */
 
 /* vss.c */
-struct vss_addr;
+struct suckaddr;
 
-int VSS_parse(const char *str, char **addr, char **port);
-int VSS_resolve(const char *addr, const char *port, struct vss_addr ***ta);
-int VSS_bind(const struct vss_addr *addr);
-int VSS_listen(const struct vss_addr *addr, int depth);
-int VSS_connect(const struct vss_addr *addr, int nonblock);
-int VSS_open(const char *str, double tmo);
+typedef int vss_resolved_f(void *priv, const struct suckaddr *);
+int VSS_resolver(const char *addr, const char *def_port, vss_resolved_f *func,
+   void *priv, const char **err);
