@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Martin Blix Grydeland <martin@varnish-software.com>
@@ -81,7 +81,20 @@
 
 #define VUT_OPT_r							\
 	VOPT("r:", "[-r filename]", "Binary file input",		\
-	    "Read log in binary file format from this file."		\
+	    "Read log in binary file format from this file. The file"	\
+	    " can be created with ``varnishlog -w filename``."		\
+	)
+
+#define VUT_OPT_t							\
+	VOPT("t:", "[-t <seconds|off>]", "VSM connection timeout",	\
+	    "Timeout before returning error on initial VSM connection."	\
+	    " If set the VSM connection is retried every 0.5 seconds"	\
+	    " for this many seconds. If zero the connection is"		\
+	    " attempted only once and will fail immediately if"		\
+	    " unsuccessful. If set to \"off\", the connection will not"	\
+	    " fail, allowing the utility to start and wait"		\
+	    " indefinetely for the Varnish instance to appear. "	\
+	    " Defaults to 5 seconds."					\
 	)
 
 #define VUT_OPT_V							\
