@@ -13,10 +13,10 @@
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
  *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
  * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -31,7 +31,8 @@
 #include "config.h"
 
 #include <stdint.h>
-#include <vct.h>
+
+#include "vct.h"
 
 /* NB: VCT always operate in ASCII, don't replace 0x0d with \r etc. */
 
@@ -72,12 +73,20 @@ const uint16_t vct_typtab[256] = {
 	[0x1e]	=	VCT_CTL,
 	[0x1f]	=	VCT_CTL,
 	[0x20]  =	VCT_SP | VCT_SEPARATOR,
+	[0x21]  =	VCT_TCHAR,
 	[0x22]  =	VCT_SEPARATOR,
+	[0x23]  =	VCT_TCHAR,
+	[0x24]  =	VCT_TCHAR,
+	[0x25]  =	VCT_TCHAR,
+	[0x26]  =	VCT_TCHAR,
+	[0x27]  =	VCT_TCHAR,
 	[0x28]  =	VCT_SEPARATOR,
 	[0x29]  =	VCT_SEPARATOR,
+	[0x2a]  =	VCT_TCHAR,
+	[0x2b]  =	VCT_TCHAR,
 	[0x2c]  =	VCT_SEPARATOR,
-	[0x2d]	=	VCT_XMLNAME,
-	[0x2e]	=	VCT_XMLNAME,
+	[0x2d]	=	VCT_XMLNAME | VCT_TCHAR,
+	[0x2e]	=	VCT_XMLNAME | VCT_TCHAR,
 	[0x2f]  =	VCT_SEPARATOR,
 	[0x30]	=	VCT_DIGIT | VCT_HEX | VCT_XMLNAME,
 	[0x31]	=	VCT_DIGIT | VCT_HEX | VCT_XMLNAME,
@@ -125,7 +134,9 @@ const uint16_t vct_typtab[256] = {
 	[0x5b]  =	VCT_SEPARATOR,
 	[0x5c]  =	VCT_SEPARATOR,
 	[0x5d]  =	VCT_SEPARATOR,
-	[0x5f]	=	VCT_XMLNAMESTART,
+	[0x5e]  =	VCT_TCHAR,
+	[0x5f]	=	VCT_XMLNAMESTART | VCT_TCHAR,
+	[0x60]	=	VCT_TCHAR,
 	[0x61]	=	VCT_LOALPHA | VCT_HEX | VCT_XMLNAMESTART,
 	[0x62]	=	VCT_LOALPHA | VCT_HEX | VCT_XMLNAMESTART,
 	[0x63]	=	VCT_LOALPHA | VCT_HEX | VCT_XMLNAMESTART,
@@ -153,7 +164,9 @@ const uint16_t vct_typtab[256] = {
 	[0x79]	=	VCT_LOALPHA | VCT_XMLNAMESTART,
 	[0x7a]	=	VCT_LOALPHA | VCT_XMLNAMESTART,
 	[0x7b]  =	VCT_SEPARATOR,
+	[0x7c]	=	VCT_TCHAR,
 	[0x7d]  =	VCT_SEPARATOR,
+	[0x7e]	=	VCT_TCHAR,
 	[0x7f]	=	VCT_CTL,
 	[0xb7]	=	VCT_XMLNAME,
 	[0xc0]	=	VCT_XMLNAMESTART,
