@@ -34,13 +34,14 @@
 
 #include <errno.h>
 #include <fcntl.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
-#include "libvarnish.h"		/* XXX: for assert() */
 #include "flopen.h"
+#include "vas.h"
 #include "vpf.h"
 
 struct vpf_fh {
