@@ -28,7 +28,6 @@
 
 #include "config.h"
 
-#include <errno.h>
 #include <pcre.h>
 #include <string.h>
 
@@ -37,12 +36,14 @@
 
 #include "vre.h"
 
-#if USE_PCRE_JIT
+#if defined(USE_PCRE_JIT)
 #define VRE_STUDY_JIT_COMPILE PCRE_STUDY_JIT_COMPILE
 #else
 #define VRE_STUDY_JIT_COMPILE 0
 #endif
 
+const unsigned VRE_has_jit = VRE_STUDY_JIT_COMPILE;
+
 #if PCRE_MAJOR < 8 || (PCRE_MAJOR == 8 && PCRE_MINOR < 20)
 #  define pcre_free_study pcre_free
 #endif
