@@ -36,6 +36,7 @@ __FBSDID("$FreeBSD: head/sys/kern/subr_vsb.c 222004 2011-05-17 06:36:32Z phk $")
 #include <stdlib.h>
 #include <string.h>
 
+#include "vdef.h"
 #include "vas.h"
 #include "vsb.h"
 
@@ -50,7 +51,6 @@ __FBSDID("$FreeBSD: head/sys/kern/subr_vsb.c 222004 2011-05-17 06:36:32Z phk $")
  */
 #define	VSB_ISDYNAMIC(s)	((s)->s_flags & VSB_DYNAMIC)
 #define	VSB_ISDYNSTRUCT(s)	((s)->s_flags & VSB_DYNSTRUCT)
-#define	VSB_ISFINISHED(s)	((s)->s_flags & VSB_FINISHED)
 #define	VSB_HASROOM(s)		((s)->s_len < (s)->s_size - 1)
 #define	VSB_FREESPACE(s)	((s)->s_size - ((s)->s_len + 1))
 #define	VSB_CANEXTEND(s)	((s)->s_flags & VSB_AUTOEXTEND)
@@ -114,7 +114,6 @@ CTASSERT(powerof2(VSB_MAXEXTENDSIZE));
 CTASSERT(powerof2(VSB_MAXEXTENDINCR));
 #endif
 
-
 static int
 VSB_extendsize(int size)
 {
@@ -143,19 +142,36 @@ VSB_extend(struct vsb *s, int addlen)
 	if (!VSB_CANEXTEND(s))
 		return (-1);
 	newsize = VSB_extendsize(s->s_size + addlen);
-	newbuf = SBMALLOC(newsize);
+	if (VSB_ISDYNAMIC(s))
+		newbuf = realloc(s->s_buf, newsize);
+	else
+		newbuf = SBMALLOC(newsize);
 	if (newbuf == NULL)
 		return (-1);
-	memcpy(newbuf, s->s_buf, s->s_size);
-	if (VSB_ISDYNAMIC(s))
-		SBFREE(s->s_buf);
-	else
+	if (!VSB_ISDYNAMIC(s)) {
+		memcpy(newbuf, s->s_buf, s->s_size);
 		VSB_SETFLAG(s, VSB_DYNAMIC);
+	}
 	s->s_buf = newbuf;
 	s->s_size = newsize;
 	return (0);
 }
 
+static void
+_vsb_indent(struct vsb *s)
+{
+	if (s->s_indent == 0 || s->s_error != 0 ||
+	    (s->s_len > 0 && s->s_buf[s->s_len - 1] != '\n'))
+		return;
+	if (VSB_FREESPACE(s) <= s->s_indent &&
+	    VSB_extend(s, s->s_indent) < 0) {
+		s->s_error = ENOMEM;
+		return;
+	}
+	memset(s->s_buf + s->s_len, ' ', s->s_indent);
+	s->s_len += s->s_indent;
+}
+
 /*
  * Initialize the internals of an vsb.
  * If buf is non-NULL, it points to a static or already-allocated string
@@ -231,29 +247,7 @@ VSB_clear(struct vsb *s)
 	VSB_CLEARFLAG(s, VSB_FINISHED);
 	s->s_error = 0;
 	s->s_len = 0;
-}
-
-/*
- * Set the vsb's end position to an arbitrary value.
- * Effectively truncates the vsb at the new position.
- */
-int
-VSB_setpos(struct vsb *s, ssize_t pos)
-{
-
-	assert_VSB_integrity(s);
-	assert_VSB_state(s, 0);
-
-	KASSERT(pos >= 0,
-	    ("attempt to seek to a negative position (%jd)", (intmax_t)pos));
-	KASSERT(pos < s->s_size,
-	    ("attempt to seek past end of vsb (%jd >= %jd)",
-	    (intmax_t)pos, (intmax_t)s->s_size));
-
-	if (pos < 0 || pos > s->s_len)
-		return (-1);
-	s->s_len = pos;
-	return (0);
+	s->s_indent = 0;
 }
 
 /*
@@ -270,6 +264,7 @@ VSB_put_byte(struct vsb *s, int c)
 
 	if (s->s_error != 0)
 		return;
+	_vsb_indent(s);
 	if (VSB_FREESPACE(s) <= 0) {
 		if (VSB_extend(s, 1) < 0)
 			s->s_error = ENOMEM;
@@ -283,38 +278,28 @@ VSB_put_byte(struct vsb *s, int c)
  * Append a byte string to an vsb.
  */
 int
-VSB_bcat(struct vsb *s, const void *buf, size_t len)
+VSB_bcat(struct vsb *s, const void *buf, ssize_t len)
 {
-	const char *str = buf;
-	const char *end = str + len;
-
 	assert_VSB_integrity(s);
 	assert_VSB_state(s, 0);
 
+	assert(len >= 0);
 	if (s->s_error != 0)
 		return (-1);
-	for (; str < end; str++) {
-		VSB_put_byte(s, *str);
+	if (len == 0)
+		return (0);
+	_vsb_indent(s);
+	if (len > VSB_FREESPACE(s)) {
+		if (VSB_extend(s, len - VSB_FREESPACE(s)) < 0)
+			s->s_error = ENOMEM;
 		if (s->s_error != 0)
 			return (-1);
 	}
+	memcpy(s->s_buf + s->s_len, buf, len);
+	s->s_len += len;
 	return (0);
 }
 
-/*
- * Copy a byte string into an vsb.
- */
-int
-VSB_bcpy(struct vsb *s, const void *buf, size_t len)
-{
-
-	assert_VSB_integrity(s);
-	assert_VSB_state(s, 0);
-
-	VSB_clear(s);
-	return (VSB_bcat(s, buf, len));
-}
-
 /*
  * Append a string to an vsb.
  */
@@ -336,20 +321,6 @@ VSB_cat(struct vsb *s, const char *str)
 	return (0);
 }
 
-/*
- * Copy a string into an vsb.
- */
-int
-VSB_cpy(struct vsb *s, const char *str)
-{
-
-	assert_VSB_integrity(s);
-	assert_VSB_state(s, 0);
-
-	VSB_clear(s);
-	return (VSB_cat(s, str));
-}
-
 /*
  * Format the given argument list and append the resulting string to an vsb.
  */
@@ -367,6 +338,7 @@ VSB_vprintf(struct vsb *s, const char *fmt, va_list ap)
 
 	if (s->s_error != 0)
 		return (-1);
+	_vsb_indent(s);
 
 	/*
 	 * For the moment, there is no way to get vsnprintf(3) to hand
@@ -383,6 +355,10 @@ VSB_vprintf(struct vsb *s, const char *fmt, va_list ap)
 		va_copy(ap_copy, ap);
 		len = vsnprintf(&s->s_buf[s->s_len], VSB_FREESPACE(s) + 1,
 		    fmt, ap_copy);
+		if (len < 0) {
+			s->s_error = errno;
+			return (-1);
+		}
 		va_end(ap_copy);
 	} while (len > VSB_FREESPACE(s) &&
 	    VSB_extend(s, len - VSB_FREESPACE(s)) == 0);
@@ -438,25 +414,6 @@ VSB_putc(struct vsb *s, int c)
 	return (0);
 }
 
-/*
- * Trim whitespace characters from end of an vsb.
- */
-int
-VSB_trim(struct vsb *s)
-{
-
-	assert_VSB_integrity(s);
-	assert_VSB_state(s, 0);
-
-	if (s->s_error != 0)
-		return (-1);
-
-	while (s->s_len > 0 && isspace(s->s_buf[s->s_len-1]))
-		--s->s_len;
-
-	return (0);
-}
-
 /*
  * Check if an vsb has an error.
  */
@@ -532,16 +489,6 @@ VSB_delete(struct vsb *s)
 		SBFREE(s);
 }
 
-/*
- * Check if an vsb has been finished.
- */
-int
-VSB_done(const struct vsb *s)
-{
-
-	return(VSB_ISFINISHED(s));
-}
-
 /*
  * Quote a string
  */
@@ -577,7 +524,10 @@ VSB_quote(struct vsb *s, const char *p, int len, int how)
 			(void)VSB_putc(s, *q);
 			break;
 		case '\n':
-			(void)VSB_cat(s, "\\n");
+			if (how & VSB_QUOTE_NONL)
+				(void)VSB_cat(s, "\n");
+			else
+				(void)VSB_cat(s, "\\n");
 			break;
 		case '\r':
 			(void)VSB_cat(s, "\\r");
@@ -597,58 +547,16 @@ VSB_quote(struct vsb *s, const char *p, int len, int how)
 }
 
 /*
- * Unquote a string
+ * Indentation
  */
-const char *
-VSB_unquote(struct vsb *s, const char *p, int len, int how)
+
+void
+VSB_indent(struct vsb * s, int i)
 {
-	const char *q;
-	char *r;
-	unsigned long u;
-	char c;
 
-	(void)how;	/* For future enhancements */
-
-	if (len == -1)
-		len = strlen(p);
-
-	for (q = p; q < p + len; q++) {
-		if (*q != '\\') {
-			(void)VSB_bcat(s, q, 1);
-			continue;
-		}
-		if (++q >= p + len)
-			return ("Incomplete '\\'-sequence at end of string");
-
-		switch(*q) {
-		case 'n':
-			(void)VSB_bcat(s, "\n", 1);
-			continue;
-		case 'r':
-			(void)VSB_bcat(s, "\r", 1);
-			continue;
-		case 't':
-			(void)VSB_bcat(s, "\t", 1);
-			continue;
-		case '0':
-		case '1':
-		case '2':
-		case '3':
-		case '4':
-		case '5':
-		case '6':
-		case '7':
-			errno = 0;
-			u = strtoul(q, &r, 8);
-			if (errno != 0 || (u & ~0xff))
-				return ("\\ooo sequence out of range");
-			c = (char)u;
-			(void)VSB_bcat(s, &c, 1);
-			q = r - 1;
-			continue;
-		default:
-			(void)VSB_bcat(s, q, 1);
-		}
-	}
-	return (NULL);
+	assert_VSB_integrity(s);
+	if (s->s_indent + i < 0)
+		s->s_error = EINVAL;
+	else
+		s->s_indent += i;
 }
