@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2014 Varnish Software AS
+ * Copyright (c) 2006-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -32,6 +32,7 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
+#include <sys/time.h>
 #ifdef HAVE_SYS_FILIO_H
 #  include <sys/filio.h>
 #endif
@@ -50,8 +51,10 @@
 #include <string.h>
 #include <unistd.h>
 
+#include "vdef.h"
 #include "vas.h"
 #include "vsa.h"
+#include "vss.h"
 #include "vtcp.h"
 
 /*--------------------------------------------------------------------*/
@@ -100,6 +103,19 @@ VTCP_name(const struct suckaddr *addr, char *abuf, unsigned alen,
 
 /*--------------------------------------------------------------------*/
 
+struct suckaddr *
+VTCP_my_suckaddr(int sock)
+{
+	struct sockaddr_storage addr_s;
+	socklen_t l;
+
+	l = sizeof addr_s;
+	AZ(getsockname(sock, (void *)&addr_s, &l));
+	return (VSA_Malloc(&addr_s, l));
+}
+
+/*--------------------------------------------------------------------*/
+
 void
 VTCP_myname(int sock, char *abuf, unsigned alen, char *pbuf, unsigned plen)
 {
@@ -212,26 +228,65 @@ VTCP_nonblocking(int sock)
  */
 
 int
-VTCP_connect(int s, const struct suckaddr *name, int msec)
+VTCP_connected(int s)
 {
-	int i, k;
+	int k;
 	socklen_t l;
+
+	/* Find out if we got a connection */
+	l = sizeof k;
+	AZ(getsockopt(s, SOL_SOCKET, SO_ERROR, &k, &l));
+
+	/* An error means no connection established */
+	errno = k;
+	if (k) {
+		AZ(close(s));
+		return (-1);
+	}
+
+	(void)VTCP_blocking(s);
+	return (s);
+}
+
+int
+VTCP_connect(const struct suckaddr *name, int msec)
+{
+	int s, i;
 	struct pollfd fds[1];
 	const struct sockaddr *sa;
 	socklen_t sl;
 
-	assert(s >= 0);
-
-	/* Set the socket non-blocking */
-	if (msec > 0)
-		(void)VTCP_nonblocking(s);
-
+	if (name == NULL)
+		return (-1);
 	/* Attempt the connect */
 	AN(VSA_Sane(name));
 	sa = VSA_Get_Sockaddr(name, &sl);
+	AN(sa);
+	AN(sl);
+
+	s = socket(sa->sa_family, SOCK_STREAM, 0);
+	if (s < 0)
+		return (s);
+
+	/* Set the socket non-blocking */
+	if (msec != 0)
+		(void)VTCP_nonblocking(s);
+
 	i = connect(s, sa, sl);
-	if (i == 0 || errno != EINPROGRESS)
-		return (i);
+	if (i == 0)
+		return (s);
+	if (errno != EINPROGRESS) {
+		AZ(close(s));
+		return (-1);
+	}
+
+	if (msec < 0) {
+		/*
+		 * Caller is responsible for waiting and
+		 * calling VTCP_connected
+		 */
+		return (s);
+	}
 
 	assert(msec > 0);
 	/* Exercise our patience, polling for write */
@@ -242,21 +297,12 @@ VTCP_connect(int s, const struct suckaddr *name, int msec)
 
 	if (i == 0) {
 		/* Timeout, close and give up */
+		AZ(close(s));
 		errno = ETIMEDOUT;
 		return (-1);
 	}
 
-	/* Find out if we got a connection */
-	l = sizeof k;
-	AZ(getsockopt(s, SOL_SOCKET, SO_ERROR, &k, &l));
-
-	/* An error means no connection established */
-	errno = k;
-	if (k)
-		return (-1);
-
-	(void)VTCP_blocking(s);
-	return (0);
+	return (VTCP_connected(s));
 }
 
 /*--------------------------------------------------------------------
@@ -278,10 +324,10 @@ VTCP_close(int *s)
 void
 VTCP_set_read_timeout(int s, double seconds)
 {
+#ifdef SO_RCVTIMEO_WORKS
 	struct timeval timeout;
 	timeout.tv_sec = (int)floor(seconds);
 	timeout.tv_usec = (int)(1e6 * (seconds - timeout.tv_sec));
-#ifdef SO_RCVTIMEO_WORKS
 	/*
 	 * Solaris bug (present at least in snv_151 and older): If this fails
 	 * with EINVAL, the socket is half-closed (SS_CANTSENDMORE) and the
@@ -292,9 +338,166 @@ VTCP_set_read_timeout(int s, double seconds)
 	    &timeout, sizeof timeout));
 #else
 	(void)s;
+	(void)seconds;
 #endif
 }
 
+/*--------------------------------------------------------------------
+ */
+
+static int __match_proto__(vss_resolved_f)
+vtcp_open_callback(void *priv, const struct suckaddr *sa)
+{
+	double *p = priv;
+
+	return (VTCP_connect(sa, (int)floor(*p * 1e3)));
+}
+
+int
+VTCP_open(const char *addr, const char *def_port, double timeout,
+    const char **errp)
+{
+	int error;
+	const char *err;
+
+	if (errp != NULL)
+		*errp = NULL;
+	assert(timeout >= 0);
+	error = VSS_resolver(addr, def_port, vtcp_open_callback,
+	    &timeout, &err);
+	if (err != NULL) {
+		if (errp != NULL)
+			*errp = err;
+		return (-1);
+	}
+	return (error);
+}
+
+/*--------------------------------------------------------------------
+ * Given a struct suckaddr, open a socket of the appropriate type, and bind
+ * it to the requested address.
+ *
+ * If the address is an IPv6 address, the IPV6_V6ONLY option is set to
+ * avoid conflicts between INADDR_ANY and IN6ADDR_ANY.
+ */
+
+int
+VTCP_bind(const struct suckaddr *sa, const char **errp)
+{
+	int sd, val, e;
+	socklen_t sl;
+	const struct sockaddr *so;
+	int proto;
+
+	if (errp != NULL)
+		*errp = NULL;
+
+	proto = VSA_Get_Proto(sa);
+	sd = socket(proto, SOCK_STREAM, 0);
+	if (sd < 0) {
+		if (errp != NULL)
+			*errp = "socket(2)";
+		return (-1);
+	}
+	val = 1;
+	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof val) != 0) {
+		if (errp != NULL)
+			*errp = "setsockopt(SO_REUSEADDR, 1)";
+		e = errno;
+		AZ(close(sd));
+		errno = e;
+		return (-1);
+	}
+#ifdef IPV6_V6ONLY
+	/* forcibly use separate sockets for IPv4 and IPv6 */
+	val = 1;
+	if (proto == AF_INET6 &&
+	    setsockopt(sd, IPPROTO_IPV6, IPV6_V6ONLY, &val, sizeof val) != 0) {
+		if (errp != NULL)
+			*errp = "setsockopt(IPV6_V6ONLY, 1)";
+		e = errno;
+		AZ(close(sd));
+		errno = e;
+		return (-1);
+	}
+#endif
+	so = VSA_Get_Sockaddr(sa, &sl);
+	if (bind(sd, so, sl) != 0) {
+		if (errp != NULL)
+			*errp = "bind(2)";
+		e = errno;
+		AZ(close(sd));
+		errno = e;
+		return (-1);
+	}
+	return (sd);
+}
+
+/*--------------------------------------------------------------------
+ * Given a struct suckaddr, open a socket of the appropriate type, bind it
+ * to the requested address, and start listening.
+ */
+
+int
+VTCP_listen(const struct suckaddr *sa, int depth, const char **errp)
+{
+	int sd;
+	int e;
+
+	if (errp != NULL)
+		*errp = NULL;
+	sd = VTCP_bind(sa, errp);
+	if (sd >= 0)  {
+		if (listen(sd, depth) != 0) {
+			e = errno;
+			AZ(close(sd));
+			errno = e;
+			if (errp != NULL)
+				*errp = "listen(2)";
+			return (-1);
+		}
+	}
+	return (sd);
+}
+
+/*--------------------------------------------------------------------*/
+
+struct helper {
+	int		depth;
+	const char	**errp;
+};
+
+static int __match_proto__(vss_resolved_f)
+vtcp_lo_cb(void *priv, const struct suckaddr *sa)
+{
+	int sock;
+	struct helper *hp = priv;
+
+	sock = VTCP_listen(sa, hp->depth, hp->errp);
+	if (sock > 0) {
+		*hp->errp = NULL;
+		return (sock);
+	}
+	AN(*hp->errp);
+	return (0);
+}
+
+int
+VTCP_listen_on(const char *addr, const char *def_port, int depth,
+    const char **errp)
+{
+	struct helper h;
+	int sock;
+
+	h.depth = depth;
+	h.errp = errp;
+
+	sock = VSS_resolver(addr, def_port, vtcp_lo_cb, &h, errp);
+	if (*errp != NULL)
+		return (-1);
+	return(sock);
+}
+
 /*--------------------------------------------------------------------
  * Set or reset SO_LINGER flag
  */
@@ -330,3 +533,51 @@ VTCP_check_hup(int sock)
 		return (1);
 	return (0);
 }
+
+/*--------------------------------------------------------------------
+ * Check if a TCP syscall return value is fatal
+ */
+
+int
+VTCP_Check(int a)
+{
+	if (a == 0)
+		return (1);
+	if (errno == ECONNRESET || errno == ENOTCONN)
+		return (1);
+#if (defined (__SVR4) && defined (__sun)) || defined (__NetBSD__)
+	/*
+	 * Solaris returns EINVAL if the other end unexpectedly reset the
+	 * connection.
+	 * This is a bug in Solaris and documented behaviour on NetBSD.
+	 */
+	if (errno == EINVAL || errno == ETIMEDOUT)
+		return (1);
+#endif
+	return (0);
+}
+
+/*--------------------------------------------------------------------
+ *
+ */
+
+int
+VTCP_read(int fd, void *ptr, size_t len, double tmo)
+{
+	struct pollfd pfd[1];
+	int i, j;
+
+	if (tmo > 0.0) {
+		pfd[0].fd = fd;
+		pfd[0].events = POLLIN;
+		pfd[0].revents = 0;
+		j = (int)floor(tmo * 1e3);
+		if (j == 0)
+			j++;
+		j = poll(pfd, 1, j);
+		if (j == 0)
+			return (-2);
+	}
+	i = read(fd, ptr, len);
+	return (i < 0 ? -1 : i);
+}
