@@ -241,7 +241,7 @@ vcc_acl_try_getaddrinfo(struct vcc *tl, struct acl_e *ae)
 /*--------------------------------------------------------------------
  * Ancient stupidity on the part of X/Open and other standards orgs
  * dictate that "192.168" be translated to 192.0.0.168.  Ever since
- * CIDR happened, "192.168/16" notation has been used, but appearantly
+ * CIDR happened, "192.168/16" notation has been used, but apparently
  * no API supports parsing this, so roll our own.
  */
 
@@ -348,13 +348,12 @@ vcc_acl_emit(struct vcc *tl, const char *acln, int anon)
 	struct acl_e *ae;
 	int depth, l, m, i;
 	unsigned at[VRT_ACL_MAXADDR + 1];
-	const char *oc;
 	struct token *t;
 	struct inifin *ifp;
 
 	Fh(tl, 0, "\nstatic int\n");
 	Fh(tl, 0,
-	    "match_acl_%s_%s(const struct vrt_ctx *ctx, const VCL_IP p)\n",
+	    "match_acl_%s_%s(VRT_CTX, const VCL_IP p)\n",
 	    anon ? "anon" : "named", acln);
 	Fh(tl, 0, "{\n");
 	Fh(tl, 0, "\tconst unsigned char *a;\n");
@@ -365,13 +364,12 @@ vcc_acl_emit(struct vcc *tl, const char *acln, int anon)
 	Fh(tl, 0, "\t\tVRT_acl_log(ctx, \"NO_FAM %s\");\n", acln);
 	Fh(tl, 0, "\t\treturn(0);\n");
 	Fh(tl, 0, "\t}\n\n");
-	if (!tl->err_unref && !anon ) {
+	if (!tl->param->err_unref && !anon ) {
 		ifp = New_IniFin(tl);
 		VSB_printf(ifp->ini,
 			"\tif (0) match_acl_named_%s(0, 0);\n", acln);
 	}
 	depth = -1;
-	oc = 0;
 	at[0] = 256;
 	VTAILQ_FOREACH(ae, &tl->acl, list) {
 
@@ -383,7 +381,6 @@ vcc_acl_emit(struct vcc *tl, const char *acln, int anon)
 		}
 
 		/* Back down, if necessary */
-		oc = "";
 		while (l <= depth) {
 			Fh(tl, 0, "\t%*s}\n", -depth, "");
 			depth--;
@@ -397,23 +394,21 @@ vcc_acl_emit(struct vcc *tl, const char *acln, int anon)
 		for (i = l; m >= 8; m -= 8, i++) {
 			if (i == 0)
 				Fh(tl, 0, "\t%*s%sif (fam == %d) {\n",
-				    -i, "", oc, ae->data[i]);
+				    -i, "", "", ae->data[i]);
 			else
 				Fh(tl, 0, "\t%*s%sif (a[%d] == %d) {\n",
-				    -i, "", oc, i - 1, ae->data[i]);
+				    -i, "", "", i - 1, ae->data[i]);
 			at[i] = ae->data[i];
 			depth = i;
-			oc = "";
 		}
 
 		if (m > 0) {
 			/* Do fractional byte compares */
 			Fh(tl, 0, "\t%*s%sif ((a[%d] & 0x%x) == %d) {\n",
-			    -i, "", oc, i - 1, (0xff00 >> m) & 0xff,
+			    -i, "", "", i - 1, (0xff00 >> m) & 0xff,
 			    ae->data[i] & ((0xff00 >> m) & 0xff));
 			at[i] = 256;
 			depth = i;
-			oc = "";
 		}
 
 		i = (ae->mask + 7) / 8;
