@@ -1,5 +1,5 @@
 /*-
- * Copyright (c) 2013 Varnish Software AS
+ * Copyright (c) 2013-2015 Varnish Software AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@FreeBSD.org>
@@ -31,7 +31,7 @@
 #include <stdlib.h>
 
 #include "cache/cache.h"
-#include "cache/cache_backend.h"
+#include "cache/cache_director.h"
 
 #include "vrt.h"
 #include "vbm.h"
@@ -51,25 +51,27 @@ vdir_expand(struct vdir *vd, unsigned n)
 }
 
 void
-vdir_new(struct vdir **vdp, const char *vcl_name, vdi_healthy *healthy,
-    vdi_getfd_f *getfd, void *priv)
+vdir_new(struct vdir **vdp, const char *name, const char *vcl_name,
+    vdi_healthy_f *healthy, vdi_resolve_f *resolve, void *priv)
 {
 	struct vdir *vd;
 
+	AN(name);
 	AN(vcl_name);
 	AN(vdp);
 	AZ(*vdp);
 	ALLOC_OBJ(vd, VDIR_MAGIC);
 	AN(vd);
 	*vdp = vd;
-	AZ(pthread_mutex_init(&vd->mtx, NULL));
+	AZ(pthread_rwlock_init(&vd->mtx, NULL));
 
 	ALLOC_OBJ(vd->dir, DIRECTOR_MAGIC);
 	AN(vd->dir);
+	vd->dir->name = name;
 	REPLACE(vd->dir->vcl_name, vcl_name);
 	vd->dir->priv = priv;
 	vd->dir->healthy = healthy;
-	vd->dir->getfd = getfd;
+	vd->dir->resolve = resolve;
 	vd->vbm = vbit_init(8);
 	AN(vd->vbm);
 }
@@ -87,7 +89,7 @@ vdir_delete(struct vdir **vdp)
 
 	free(vd->backend);
 	free(vd->weight);
-	AZ(pthread_mutex_destroy(&vd->mtx));
+	AZ(pthread_rwlock_destroy(&vd->mtx));
 	free(vd->dir->vcl_name);
 	FREE_OBJ(vd->dir);
 	vbit_destroy(vd->vbm);
@@ -95,17 +97,24 @@ vdir_delete(struct vdir **vdp)
 }
 
 void
-vdir_lock(struct vdir *vd)
+vdir_rdlock(struct vdir *vd)
 {
 	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
-	AZ(pthread_mutex_lock(&vd->mtx));
+	AZ(pthread_rwlock_rdlock(&vd->mtx));
+}
+
+void
+vdir_wrlock(struct vdir *vd)
+{
+	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
+	AZ(pthread_rwlock_wrlock(&vd->mtx));
 }
 
 void
 vdir_unlock(struct vdir *vd)
 {
 	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
-	AZ(pthread_mutex_unlock(&vd->mtx));
+	AZ(pthread_rwlock_unlock(&vd->mtx));
 }
 
 
@@ -116,7 +125,7 @@ vdir_add_backend(struct vdir *vd, VCL_BACKEND be, double weight)
 
 	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
 	AN(be);
-	vdir_lock(vd);
+	vdir_wrlock(vd);
 	if (vd->n_backend >= vd->l_backend)
 		vdir_expand(vd, vd->l_backend + 16);
 	assert(vd->n_backend < vd->l_backend);
@@ -129,7 +138,33 @@ vdir_add_backend(struct vdir *vd, VCL_BACKEND be, double weight)
 }
 
 unsigned
-vdir_any_healthy(struct vdir *vd, double *changed)
+vdir_remove_backend(struct vdir *vd, VCL_BACKEND be)
+{
+	unsigned u, n;
+
+	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
+	if (be == NULL)
+		return (vd->n_backend);
+	CHECK_OBJ(be, DIRECTOR_MAGIC);
+	vdir_wrlock(vd);
+	for (u = 0; u < vd->n_backend; u++)
+		if (vd->backend[u] == be)
+			break;
+	if (u == vd->n_backend) {
+		vdir_unlock(vd);
+		return (vd->n_backend);
+	}
+	vd->total_weight -= vd->weight[u];
+	n = (vd->n_backend - u) - 1;
+	memmove(&vd->backend[u], &vd->backend[u+1], n * sizeof(vd->backend[0]));
+	memmove(&vd->weight[u], &vd->weight[u+1], n * sizeof(vd->weight[0]));
+	vd->n_backend--;
+	vdir_unlock(vd);
+	return (vd->n_backend);
+}
+
+unsigned
+vdir_any_healthy(struct vdir *vd, const struct busyobj *bo, double *changed)
 {
 	unsigned retval = 0;
 	VCL_BACKEND be;
@@ -137,13 +172,14 @@ vdir_any_healthy(struct vdir *vd, double *changed)
 	double c;
 
 	CHECK_OBJ_NOTNULL(vd, VDIR_MAGIC);
-	vdir_lock(vd);
+	CHECK_OBJ_ORNULL(bo, BUSYOBJ_MAGIC);
+	vdir_rdlock(vd);
 	if (changed != NULL)
 		*changed = 0;
 	for (u = 0; u < vd->n_backend; u++) {
 		be = vd->backend[u];
 		CHECK_OBJ_NOTNULL(be, DIRECTOR_MAGIC);
-		retval = be->healthy(be, &c);
+		retval = be->healthy(be, bo, &c);
 		if (changed != NULL && c > *changed)
 			*changed = c;
 		if (retval)
@@ -174,33 +210,25 @@ vdir_pick_by_weight(const struct vdir *vd, double w,
 }
 
 VCL_BACKEND
-vdir_pick_be(struct vdir *vd, double w, unsigned nloops)
+vdir_pick_be(struct vdir *vd, double w, const struct busyobj *bo)
 {
-	struct vbitmap *vbm = NULL;
-	unsigned u, v, l;
+	unsigned u;
+	double tw = 0.0;
 	VCL_BACKEND be = NULL;
-	double tw;
-	int nbe;
 
-	tw = vd->total_weight;
-	nbe = vd->n_backend;
-	assert(w >= 0.0 && w < 1.0);
-	vdir_lock(vd);
-	for (l = 0; nbe > 0 && tw > 0.0 && l < nloops; l++) {
-		u = vdir_pick_by_weight(vd, w * tw, vbm);
+	vdir_rdlock(vd);
+	for (u = 0; u < vd->n_backend; u++) {
+		if (vd->backend[u]->healthy(vd->backend[u], bo, NULL)) {
+			vbit_clr(vd->vbm, u);
+			tw += vd->weight[u];
+		} else
+			vbit_set(vd->vbm, u);
+	}
+	if (tw > 0.0) {
+		u = vdir_pick_by_weight(vd, w * tw, vd->vbm);
+		assert(u < vd->n_backend);
 		be = vd->backend[u];
 		CHECK_OBJ_NOTNULL(be, DIRECTOR_MAGIC);
-		if (be->healthy(be, NULL))
-			break;
-		if (l == 0) {
-			vbm = vd->vbm;
-			for (v = 0; v < nbe; v++)
-				vbit_clr(vbm, v);
-		}
-		vbit_set(vbm, u);
-		nbe--;
-		tw -= vd->weight[u];
-		be = NULL;
 	}
 	vdir_unlock(vd);
 	return (be);
